package fi.jyu.mit.ohj2;
import java.util.Enumeration;

/**
 * Luokka StringTokenizerin korvaajaksi.  Erona
 * on se ett jonon loppuessa ei tule ongelmia ja 
 * perkkisten erottimien vlist tulee tyhj jono.
 * <pre>
 * Esimerkki:
 * public static void main(String[] args) {
 *     Erottelija erottaja = new Erottelija("12;3.5:kissa,,,istuu puussa,3.4",
 *                                          ";:,");
 *     System.out.println("Palasia: " + erottaja.countTokens());
 *     for (int i=1; erottaja.hasMoreTokens(); i++ )
 *       System.out.println(i + ": |" + erottaja.nextToken()+"|");
 *     System.err.println("8: |"+erottaja.nextToken()+"|");
 *     erottaja.reset();
 *     System.out.println(erottaja.nextToken(0));
 *     System.out.println(erottaja.countRemaininTokens());
 *     System.out.println(erottaja.rest());
 *     System.out.println(erottaja.nextToken(0.0));
 *     System.out.println(erottaja.nextToken(2));
 *     System.out.println(erottaja.nextToken(2.1));
 *     System.out.println(erottaja.countRemainingTokens());
 *     System.out.println(erottaja.rest());
 *   }
 *
 * Tulostaa:
 * 
 * Palasia: 7
 * 1: |12|
 * 2: |3.5|
 * 3: |kissa|
 * 4: ||
 * 5: ||
 * 6: |istuu puussa|
 * 7: |3.4|
 * 8: ||
 * 12
 * 6
 * 3.5:kissa,,,istuu puussa,3.4
 * 3
 * ,istuu puussa,3.4
 * 12
 * 3.5
 * 2
 * 2.1
 * </pre>
 * @author vesal
 * @version 11.3.2007
 * 
 * @example
 * <pre name="testErottelija">
 * Erottelija erottaja = new Erottelija("12;3.5:kissa,,,istuu puussa,3.4",
 *                                          ";:,");
 * erottaja.nextToken() === $s; erottaja.hasMoreTokens() === $tokens;
 * erottaja.countRemainingTokens() === $n; erottaja.rest() === $rest;   
 * 
 *   $i  |  $s             | $tokens | $n | $rest
 *  --------------------------------------------------------------------------  
 *   0   | ---             | true    | 7  | "12;3.5:kissa,,,istuu puussa,3.4"
 *   1   | "12"            | true    | 6  | "3.5:kissa,,,istuu puussa,3.4"
 *   2   | "3.5"           | true    | 5  | "kissa,,,istuu puussa,3.4"
 *   3   | "kissa"         | true    | 4  | ",,istuu puussa,3.4"
 *   4   | ""              | true    | 3  | ",istuu puussa,3.4"
 *   5   | ""              | true    | 2  | "istuu puussa,3.4"
 *   6   | "istuu puussa"  | true    | 1  | "3.4"
 *   7   | "3.4"           | false   | 0  | ""
 *   8   | ""              | false   | 0  | ""
 *   9   | ""              | false   | 0  | ""
 *   
 *   erottaja.nextToken(";","kissa") === "kissa";
 * 
 * @example
 * </pre>
 * <pre name="testErottelijaEnd">
 * Erottelija erottaja = new Erottelija("12;3.5:kissa,,,istuu puussa,3.4;;",
 *                                          ";:,");
 * erottaja.nextToken() === $s; erottaja.hasMoreTokens() === $tokens;
 * erottaja.countRemainingTokens() === $n; erottaja.rest() === $rest;   
 * 
 *   $i  |  $s             | $tokens | $n | $rest
 *  --------------------------------------------------------------------------  
 *   0   | ---             | true    | 9  | "12;3.5:kissa,,,istuu puussa,3.4;;"
 *   1   | "12"            | true    | 8  | "3.5:kissa,,,istuu puussa,3.4;;"
 *   2   | "3.5"           | true    | 7  | "kissa,,,istuu puussa,3.4;;"
 *   3   | "kissa"         | true    | 6  | ",,istuu puussa,3.4;;"
 *   4   | ""              | true    | 5  | ",istuu puussa,3.4;;"
 *   5   | ""              | true    | 4  | "istuu puussa,3.4;;"
 *   6   | "istuu puussa"  | true    | 3  | "3.4;;"
 *   7   | "3.4"           | true    | 2  | ";"
 *   8   | ""              | true    | 1  | ""
 *   9   | ""              | false   | 0  | ""
 *  10   | ""              | false   | 0  | ""
 *   
 *   erottaja.nextToken(";","kissa") === "kissa";
 * </pre>
 */
@SuppressWarnings("unchecked")
public class Erottelija implements Enumeration { // NOPMD - enum ok

  /**
   * Etsitn mist kohti jonosta str lytyy ensimminen erotinmerkki 
   * joukosta delim.  Etsint aloitetaan paikasta pos.
   * @param str    mist jonosta etsitn
   * @param delim  joukko erotinmerkkej
   * @param pos    paikka josta aloitetaan 
   * @return       palauttaa ensimmisen esiintymn tai -1 jos ei lydy
   * @example
   * <pre name="test">
   *   indexOfAny("a;, b",",; ",0) === 1
   *   indexOfAny("a;, b",",; ",2) === 2
   *   indexOfAny("a;, b"," ",0)   === 3
   *   indexOfAny("a;, b",".",0)   === -1
   *   indexOfAny(null,",; ",0)    === -1
   *   indexOfAny("a b",",; ",-1)  === 1
   * </pre>
   */
  public static int indexOfAny(String str, String delim, int pos) {
    int i=pos; if ( i<0 ) i = 0;
    if ( str == null || delim == null ) return -1;
    for ( ; i<str.length(); i++ ) {
      char c = str.charAt(i);
      if ( delim.indexOf(c) >= 0 ) return i;
    }
    return -1;
  }

  /**
   * Etsitn mist kohti jonosta str lytyy ensimminen erotinmerkki 
   * joukosta delim.  Etsint aloitetaan alusta.
   * @param str    mist jonosta etsitn
   * @param delim  joukko erotinmerkkej
   * @return       palauttaa ensimmisen esiintymn tai -1 jos ei lydy
   * @example
   * <pre name="test">
   *   indexOfAny("a;, b",",; ") === 1
   * </pre>
   */
  public static int indexOfAny(String str, String delim) {
    return indexOfAny(str,delim,0);
  }


  private String str;
  private final String delim;
  private int pos=0;

  /**
   * Luodaan erottelija, joka erottelee jonosta str palasia mink tahansa
   * joukosta delim lytyvn merkin kohdalta. 
   * @param str    jono josta erotellaan
   * @param delim  erottavien merkkien joukko
   */
  public Erottelija(String str, String delim) {
    this.str = str;
    this.delim = delim;
  }

  /**
   * Luodaan erottelija, joka erottelee jonosta str palasia vlilynnin 
   * kohdalta.
   * @param str    jono josta erotellaan
   * @example
   * <pre name="test">
   *                                     //01234 
   * Erottelija erottaja = new Erottelija("a b ");
   * erottaja.countTokens() === 3;
   * </pre>
   */
  public Erottelija(String str) {
    this.str = str; 
    this.delim = " ";
  }

  /**
   * Palauttaa seuraavan palasen jonosta
   * @return jonon seuraava palanen
   * @example
   * <pre name="test">
   * Erottelija erottaja = new Erottelija("a;b;",";");
   * erottaja.nextToken() === $s; erottaja.hasMoreTokens() === $tokens;
   * erottaja.countRemainingTokens() === $n; erottaja.rest() === $rest;   
   * 
   *   $i  |  $s             | $tokens | $n | $rest
   *  --------------------------------------------------------------------------  
   *   0   | ---             | true    | 3  | "a;b;"
   *   1   | "a"             | true    | 2  | "b;"
   *   2   | "b"             | true    | 1  | ""
   *   3   | ""              | false   | 0  | ""
   *   4   | ""              | false   | 0  | ""
   * </pre>
   */
  public String nextToken() {
    return nextToken(delim);
  }

  /**
   * Tarkistaa onko erotinmerkki juuri paikan vasemmalla puolella
   * @param pos paikka josta tutkitaan
   * @param delim kelpaavat erotinmerkit
   * @return true jos on erotin merkki vasemmalla puolella, false muuten
   * @example
   * <pre name="test">
   *                                     //01234 
   * Erottelija erottaja = new Erottelija("a;b;",";");
   * erottaja.isDelimBefore(0,";") === false;
   * erottaja.isDelimBefore(1,";") === false;
   * erottaja.isDelimBefore(2,";") === true;
   * erottaja.isDelimBefore(3,";") === false;
   * erottaja.isDelimBefore(4,";") === true;
   * erottaja.isDelimBefore(5,";") === false;
   * </pre>
   */
  public boolean isDelimBefore(int pos, String delim) {
     if ( pos <= 0 ) return false;
     if ( pos > str.length() ) return false;
     String usedDelim = delim;
     if ( delim == null ) usedDelim = this.delim;
     char c = str.charAt(pos-1);
     return ( usedDelim.indexOf(c) >= 0 );
  }
  
  /**
   * Tarkistaa onko erotinmerkki juuri paikan vasemmalla puolella
   * @param pos paikka josta tutkitaan
   * @return true jos on erotin merkki vasemmalla puolella, false muuten
   * @example
   * <pre name="test">
   *                                     //01234 
   * Erottelija erottaja = new Erottelija("a;b;",";");
   * erottaja.isDelimBefore(0) === false;
   * erottaja.isDelimBefore(1) === false;
   * erottaja.isDelimBefore(2) === true;
   * erottaja.isDelimBefore(3) === false;
   * erottaja.isDelimBefore(4) === true;
   * erottaja.isDelimBefore(5) === false;
   * </pre>
   */
  public boolean isDelimBefore(int pos) {
     return isDelimBefore(pos,null);
  }
  
  /**
   * Tarkistaa onko erotinmerkki juuri nykypaikan vasemmalla puolella
   * @return true jos on erotin merkki vasemmalla puolella, false muuten
   * @example
   * <pre name="test">
   *                                     //01234 
   * Erottelija erottaja = new Erottelija("a;b;",";");
   *                               erottaja.isDelimBefore() === false;
   * erottaja.nextToken() === "a"; erottaja.isDelimBefore() === true;
   * erottaja.nextToken() === "b"; erottaja.isDelimBefore() === true;
   * erottaja.nextToken() === "";  erottaja.isDelimBefore() === false;
   * erottaja.nextToken() === "";  erottaja.isDelimBefore() === false;
   * </pre>
   */
  public boolean isDelimBefore() {
     return isDelimBefore(pos,null);
  }
  
  
  /**
   * Palauttaa seuraavan palasen jonosta.
   * @param delim erotinjoukko, jonka perusteella perusteella erotetaan
   * @return jonon seuraava palanen
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("a b;c");
   *   erottaja.nextToken(" ") === "a";
   *   erottaja.nextToken(" ") === "b;c";
   *   erottaja = new Erottelija("a b;c");
   *   erottaja.nextToken(" ") === "a";
   *   erottaja.nextToken(";") === "b";
   *   erottaja.nextToken(" ") === "c";
   *   erottaja.nextToken(" ") === "";
   *   erottaja = new Erottelija(null);
   *   erottaja.nextToken(" ") === "";
   *   erottaja = new Erottelija("a b");
   *   erottaja.nextToken(null) === "a";
   * </pre>
   */
  public String nextToken(String delim) {
    if ( str == null ) return "";
    int len = str.length();
    if ( pos > len ) return "";
    if ( pos == len ) { pos = len+1; return ""; }
    String usedDelim = delim;
    if ( delim == null ) usedDelim = this.delim;
    int nextpos = indexOfAny(str,usedDelim,pos);
    if ( nextpos < 0 ) nextpos = len;
    String result = str.substring(pos,nextpos);
    pos = nextpos;
    if ( pos < len ) pos++;
    return result;
  }

  /**
   * Ottaa seuraavan palasen ja jos se on tyhj, niin palauttaa def-jonon.
   * @param delim erotinjoukko, jonka perusteella perusteella erotetaan
   * @param def oletusarvo jos seuraava palanen on tyhj
   * @return jonon seuraava palanen tai oletus
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("a b;c");
   *   erottaja.nextToken(" ","d") === "a";
   *   erottaja.nextToken(" ","d") === "b;c";
   *   erottaja.nextToken(" ","d") === "d";
   * </pre>  
   */
  public String nextToken(String delim, String def) {
    String piece = nextToken(delim);
    if ( piece.length() > 0 ) return piece;
    return def;
  }

  /**
   * Palauttaa jonosta seuraavan kokonaisluvun ja oletuksen jos luku ei ole
   * kunnollinen.
   * @param delim erotinjoukko, jonka perusteella perusteella erotetaan
   * @param def oletusarvo jos luku ei ole kunnollinen
   * @return seuraava kokonaisluku tai oletus
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1;2");
   *   erottaja.nextToken(";",3) === 1; 
   *   erottaja.nextToken(";",3) === 2; 
   *   erottaja.nextToken(";",3) === 3; 
   * </pre>
   */
  public int nextToken(String delim, int def) {
    String piece = nextToken(delim);
    return Mjonot.erotaInt(piece,def);
  }

  /**
   * Palauttaa jonosta seuraavan kokonaisluvun ja oletuksen jos luku ei ole
   * kunnollinen.
   * @param def oletusarvo jos luku ei ole kunnollinen
   * @return seuraava kokonaisluku tai oletus
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1 2");
   *   erottaja.nextToken(3) === 1; 
   *   erottaja.nextToken(3) === 2; 
   *   erottaja.nextToken(3) === 3; 
   * </pre>
   */
  public int nextToken(int def) {
    return nextToken(null,def);
  }

  
  /**
   * Palauttaa jonosta seuraavan kokonaisluvun ja 0 jos luku ei ole
   * kunnollinen.
   * @return seuraava kokonaisluku tai 0
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1 2");
   *   erottaja.nextInt() === 1; 
   *   erottaja.nextInt() === 2; 
   *   erottaja.nextInt() === 0; 
   * </pre>
   */
  public int nextInt() {
    return nextToken(0);
  }
  
  
  /**
   * Palauttaa jonosta seuraavan reaaliluvun ja oletuksen jos luku ei ole
   * kunnollinen.
   * @param delim erotinjoukko, jonka perusteella perusteella erotetaan
   * @param def oletusarvo jos luku ei ole kunnollinen
   * @return seuraava reaaliluku tai oletus
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1;2");
   *   erottaja.nextToken(";",3.1) ~~~ 1.0; 
   *   erottaja.nextToken(";",3.1) ~~~ 2.0; 
   *   erottaja.nextToken(";",3.1) ~~~ 3.1; 
   * </pre>
   */
  public double nextToken(String delim, double def) {
    String piece = nextToken(delim);
    return Mjonot.erotaDouble(piece,def);
  }

  /**
   * Palauttaa jonosta seuraavan reaaliluvun ja oletuksen jos luku ei ole
   * kunnollinen.
   * @param def oletusarvo jos luku ei ole kunnollinen
   * @return seuraava reaaliluku tai oletus
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1 2");
   *   erottaja.nextToken(3.1) ~~~ 1.0; 
   *   erottaja.nextToken(3.1) ~~~ 2.0; 
   *   erottaja.nextToken(3.1) ~~~ 3.1; 
   * </pre>
   */
  public double nextToken(double def) {
    return nextToken(null,def);
  }

  /**
   * Palauttaa jonosta seuraavan reaaliluvun ja 0.0 jos luku ei ole
   * kunnollinen.
   * @return seuraava reaaliluku tai 0.0
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1 2");
   *   erottaja.nextDouble() ~~~ 1.0; 
   *   erottaja.nextDouble() ~~~ 2.0; 
   *   erottaja.nextDouble() ~~~ 0.0; 
   * </pre>
   */
  public double nextDouble() {
    return nextToken(0.0);
  }
  
  
  /**
   * Laskee palasten lukumrn.
   * @param pos paikka josta laskeminen aloitetaan
   * @return palasten lukumr.
   */
  public int countTokens(int pos) {
    int n=1;
    int len=str.length();
    if ( pos > len ) return 0;
    if ( pos == len ) return isDelimBefore(pos) ? 1 : 0;
    for (int i=pos; i<len; i++) {
      char c = str.charAt(i);
      if ( delim.indexOf(c) >= 0 ) n++;
    }
    return n;
  }

  /**
   * Laskee palasten lukumrn.
   * @return palasten lukumr.
   */
  public int countTokens() {
    return countTokens(0);
  }
   
  
  /**
   * Laskee palasten lukumrn.
   * @return palasten lukumr.
   */
  public int countRemainingTokens() {
    return countTokens(pos);
  }

  /**
   * Tarkistaa ett vielk palasia on jljell.
   * @return onko palasia jljell
   */
  public boolean hasMoreElements() {
    if ( isDelimBefore() ) return true;  
    return ( pos < str.length() );
  }

  /**
   * Tarkistaa ett vielk palasia on jljell.
   * @return onko palasia jljell
   */
  public boolean hasMoreTokens() {
    return hasMoreElements();
  }

  /**
   * Palauttaa seuraavan palasen Objectina.
   * @return seuraava palanen
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija("1 2");
   *   erottaja.nextElement().toString() === "1";
   * </pre>
   */
  public Object nextElement() {
    return nextToken();
  }

  /**
   * Siivoaa palasteltavan jonon turhista vlilynneist
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija(" 1   2 ");
   *   erottaja.countTokens() === 6;
   *   erottaja.trim();
   *   erottaja.countTokens() === 4;
   * </pre>
   */
  public void trim() {
    str = Mjonot.poista_2_tyhjat(str);
    pos = 0;
  }
  
  /**
   * Palauttaa jljell olevan jonon.
   * @return jljell oleva jono.
   * @example
   * <pre name="test">
   *   Erottelija erottaja = new Erottelija(" 1   2 ");
   *   erottaja.trim();
   *   erottaja.rest() === " 1 2 ";
   * </pre>
   */
  public String rest() {
    if ( !hasMoreTokens() ) return "";
    return str.substring(pos);
  }
  
  /**
   * Palauttaa paikan erottelijan alkuun.
   */
  public void reset() {
    pos = 0;    
  }

  /**
   * Testataan Erottelijaluokkaa
   * @param args ei kytss
   */
/*
  public static void main(String[] args) {
    Erottelija erottaja = new Erottelija("12;3.5:kissa,,,istuu puussa,3.4",";:,");
    System.out.println("Palasia: " + erottaja.countTokens());
    for (int i=1; erottaja.hasMoreTokens(); i++ )
      System.out.println(i + ": |" + erottaja.nextToken()+"|");
    System.out.println("8: |"+erottaja.nextToken()+"|");
    erottaja.reset();
    System.out.println(erottaja.nextToken(0));
    System.out.println(erottaja.countRemainingTokens());
    System.out.println(erottaja.rest());
    System.out.println(erottaja.nextToken(0.0));
    System.out.println(erottaja.nextToken(2));
    System.out.println(erottaja.nextToken(2.1));
    System.out.println(erottaja.countRemainingTokens());
    System.out.println(erottaja.rest());
  }
*/
}
