package fi.jyu.mit.ohj2;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;

/**
 * Joukko merkkijonoaliohjelmia Ohjelmointi 2-kurssin kyttn
 * @author Vesa Lappalainen
 * @version 1.0, 07.01.2003
 * @version 1.1, 09.03.2002
 * @version 1.2, 30.09.2007 / vl, testi
 * @version 1.3, 12.08.2008 / vl, StringBuilder
 * @since 1.5
 */

public class Mjonot { // NOPMD

  //****************************************************************************
    
  /**
   * Palauttaa jonon, jossa on monistettuna jono s n-kertaa
   * <pre> tayta("kissa",3)  => "kissakissakissa"</pre>
   * Tuplataan jonon pituus joka kerta => 100 monistusta vaatii 7 kierrosta
   * @param s monistettava jono
   * @param n kuinka monta kertaa monistetaan
   * @return s monistettuna n kertaa
   * 
   * @example
   * <pre name="testTayta">
   * tayta($jono,$n) === $tulos
   * 
   *    $jono   | $n | $tulos
   *  -------------------------------
   *    "kissa" |  3 | "kissakissakissa"
   *    ""      | 10 | ""
   *    "kissa" |  0 | ""     
   *    "kissa" |  1 | "kissa"
   *    "a"     | 10 | "aaaaaaaaaa"     
   * </pre>
   */
  public static String tayta(String s, int n) {
    int pituus = n*s.length();
    if ( pituus == 0 ) return "";
    StringBuilder jono = new StringBuilder(pituus);
    jono.append(s);
    while ( 2*jono.length() < pituus  ) jono.append(jono);
    jono.append(jono.substring(0,pituus-jono.length()));
    return jono.toString();
  }


  /** Palauttaa merkkijonon jossa on n kappaletta vlilyntej
   * @param n montako vlilynti tarvitaan
   * @return jono, jossa on n kpl vlilyntej
   * 
   *  @example
   * <pre name="test">
   * tyhja(0) === ""
   * tyhja(1) === " "
   * tyhja(2) === "  "
   * tyhja(10) === "          ";
   * tyhja(1000).length() === 1000; 
   * </pre>
   */
  public static String tyhja(int n) {
    StringBuilder fill = new StringBuilder(40+n); // NOPMD sopivan kokoinen
    while ( fill.length() < n ) fill.append("                                ");
    fill.delete(n,fill.length());
    return fill.toString();
  }

  
  /**
   * Palauttaa jonon s muotoiltuna vhintn len-pituiseksi
   * <pre>
   * Esim:  fmt("2",3,'0')  => "002"
   *        fmt("2",-3,' ') => "2  "
   * </pre>
   * @param s      muotoiltava jono
   * @param len    pituus, negatiivisella vasempaan laitaan, pos. oikeaan
   * @param fillch merkki jolla tytetn
   * @return       muotoiltu jono
   * 
   * @example
   * <pre name="test">
   *   fmt("",3,' ')     === "   "  
   *   fmt("",3,'0')     === "000"  
   *   fmt("2",3,'0')    === "002"  
   *   fmt("2",-3,' ')   === "2  "
   *   fmt("1234",3,'0') === "1234" 
   * </pre>
   */
  public static String fmt(String s,int len,char fillch) {
    int needs = Math.abs(len) - s.length();
    if ( needs <= 0 ) return s;
    String fill = tayta(""+fillch,needs);
    if ( len < 0 ) return s + fill;
    return fill + s;
  }

  
  /**
   * Palauttaa jonon s muotoiltuna vhintn len-pituiseksi
   * <pre>
   * Esim:  fmt("2",3)  => "  2"
   *        fmt("2",-3) => "2  "
   * </pre>
   * @param s     muotoiltava jono
   * @param len   pituus, negatiivisella vasempaan laitaan, pos. oikeaan
   * @return      muotoiltu jono
   * 
   * @example
   * <pre name="test">
   *   fmt("2",3)    === "  2"  
   *   fmt("2",-3)   === "2  "
   *   fmt("1234",3) === "1234" 
   * </pre>
   */
  public static String fmt(String s,int len) {
    int needs = Math.abs(len) - s.length();
    if ( needs <= 0 ) return s;
    String fill = tyhja(needs);
    if ( len < 0 ) return s + fill;
    return fill + s;
  }

  
  /**
   * Palauttaa jonon s muotoiltuna vhintn len-pituiseksi
   * @param s     muotoiltava jono
   * @param len   pituus, negatiivisella vasempaan laitaan, pos. oikeaan
   * @return      muotoiltu jono
   * @see #fmt(String s,int len)
   * @example
   * <pre name="test">
   *   fmt(new StringBuffer("12"),3) === " 12"
   * </pre>
   */
  public static String fmt(StringBuffer s,int len) {
    return fmt(s.toString(),len);
  }

  
  /**
   * Palauttaa jonon s muotoiltuna vhintn len-pituiseksi
   * @param s     muotoiltava jono
   * @param len   pituus, negatiivisella vasempaan laitaan, pos. oikeaan
   * @return      muotoiltu jono
   * @see #fmt(String s,int len)
   * @example
   * <pre name="test">
   *   fmt(new StringBuilder("12"),3) === " 12"
   * </pre>
   */
  public static String fmt(StringBuilder s,int len) {
    return fmt(s.toString(),len);
  }

  
  /**
   * Muotoilee kokonaisluvun len kokoiseen tilaan.
   * <pre>fmt(12,4) => "  12";  fmt(12,-4) => "12  "</pre>
   *
   * @param n    muotoiltava luku
   * @param len  tulosjonon vhimmispituus
   * @return     merkkijonoksi muotoiltu luku
   * @see #fmt(String s,int len)
   * 
   * @example
   * <pre name="test">
   *   fmt(2,3)    === "  2"  
   *   fmt(2,-3)   === "2  "
   *   fmt(1234,3) === "1234" 
   * </pre>
   */
  public static String fmt(int n,int len) {
    return fmt(""+n,len);
  }

  
  /**
   * Muotoilee kokonaisluvun len kokoiseen tilaan.
   * <pre>fmt(12,4,'0') => "0012";  fmt(12,-4,' ') => "12  "</pre>
   *
   * @param n      muotoiltava luku
   * @param len    tulosjonon vhimmispituus
   * @param fillch merkki jolla tytetn
   * @return       merkkijonoksi muotoiltu luku
   * @see #fmt(String s,int len)
   * 
   * @example
   * <pre name="test">
   *   fmt(2,3,'0')    === "002"  
   *   fmt(2,-3,'0')   === "200"
   *   fmt(2,-3,' ')   === "2  "
   *   fmt(1234,3,'0') === "1234"
   *   fmt(12,4,'0')   === "0012"  
   *   fmt(12,-4,' ')  === "12  " 
   * </pre>
   */
  public static String fmt(int n,int len, char fillch) {
    return fmt(""+n,len,fillch);
  }


  /**
   * Muotoilee reaaliluvun len kokoiseen tilaan.
   * <pre>fmt(12.3,6,2) => " 12.30";  fmt(12.3,-6,2) => "12.30 "</pre>
   * @param d     muotoiltava luku
   * @param len   tulosjonon vhimmispituus
   * @param desim desimaalien lukumr tulosjonossa
   * @return      merkkijonoksi muotoiltu luku
   * @see #fmt(String s,int len)
   * 
   * @example
   * <pre name="test">
   *   fmt(12.3,6,2)  === " 12.30"  
   *   fmt(12.3,-6,2) === "12.30 "
   *   fmt(2.0,-6,2)  === "2.00  "
   *   fmt(1234,6,0)  === "  1234"
   *   fmt(12,-4,0)   === "12  " 
   *   fmt(12.1,-4,0) === "12  " 
   * </pre>
   * 
   */
  public static String fmt(double d,int len,int desim) {
    String des = "0";
    if ( desim > 0 ) des = "0."+tayta("0",desim);
    DecimalFormat df = new DecimalFormat(des);
    DecimalFormatSymbols ds = new DecimalFormatSymbols();
    ds.setDecimalSeparator('.');
    df.setDecimalFormatSymbols(ds);
    return fmt(df.format(d),len);
  }


  /**
   * Poistetaan jonosta kaikki kaksin- tai useampikertaiset tyhjt.
   * <pre>
   * jono = "   kissa     istuu   "; 
   * poista_2_tyhjat(jono)  => jono = " kissa istuu ";
   * </pre>
   * @param s jono josta tyhjt poistetaan
   * @return viite jonoon josta poistettu
   * @example
   * <pre name="test">
   *   StringBuffer sb = new StringBuffer("   kissa     istuu   ");
   *   poista_2_tyhjat(sb).toString() === " kissa istuu ";
   *   poista_2_tyhjat((StringBuffer)null).toString() === "";
   * </pre>
   * 
   */
  public static StringBuffer poista_2_tyhjat(StringBuffer s) {  // NOPMD sallitaan alleviiva historian painolastina
    int tyhjia=0;
    int l=0,k=0;
    if ( s == null ) return new StringBuffer("");
    int pit = s.length();

    while ( l < pit ) {
      char c = s.charAt(l);
      if ( c == ' ') tyhjia++;
      else tyhjia = 0;
      if ( tyhjia <= 1 ) s.setCharAt(k++,c);  // Jollei 2-tyhj, kopioidaan!
      l++;
    }
    s.delete(k,pit);
    return s;
  }

  
  /**
   * Poistetaan jonosta kaikki kaksin- tai useampikertaiset tyhjt.
   * <pre>
   * jono = "   kissa     istuu   "; 
   * poista_2_tyhjat(jono)  => jono = " kissa istuu ";
   * </pre>
   * @param s jono josta tyhjt poistetaan
   * @return viite jonoon josta poistettu
   * @example
   * <pre name="test">
   *   StringBuilder sb = new StringBuilder("   kissa     istuu   ");
   *   poista_2_tyhjat(sb).toString() === " kissa istuu ";
   *   poista_2_tyhjat((StringBuilder)null).toString() === "";
   * </pre>
   * 
   */
  public static StringBuilder poista_2_tyhjat(StringBuilder s) {  // NOPMD sallitaan alleviiva historian painolastina
    int tyhjia=0;
    int l=0,k=0;
    if ( s == null ) return new StringBuilder("");
    int pit = s.length();

    while ( l < pit ) {
      char c = s.charAt(l);
      if ( c == ' ') tyhjia++;
      else tyhjia = 0;
      if ( tyhjia <= 1 ) s.setCharAt(k++,c);  // Jollei 2-tyhj, kopioidaan!
      l++;
    }
    s.delete(k,pit);
    return s;
  }

  
  /**
   * Poistetaan jonosta kaikki kaksin- tai useampikertaiset tyhjt.
   * <pre>
   * jono = poista_2_tyhjat("   kissa     istuu   ")  => jono = " kissa istuu ";
   * </pre>
   * @param s jono josta tyhjt poistetaan
   * @return viite uusi jono josta tyhjt poistettu
   * 
   * @example
   * <pre name="test">
   * poista_2_tyhjat("   kissa     istuu   ") === " kissa istuu "  
   * poista_2_tyhjat("") === ""  
   * poista_2_tyhjat("   ") === " "  
   * poista_2_tyhjat((String)null) === "";
   * </pre>
   */
  public static String poista_2_tyhjat(String s) { // NOPMD sallitaan alleviiva historian painolastina
    if ( s == null ) return "";
    return poista_2_tyhjat(new StringBuilder(s)).toString();
  }


  //****************************************************************************
  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * Alkuosa palautetaan funktion nimess ja loppuosa jtetn
   * jonoon.  Merkin etsimissuunta voidaan valita (oletuksena alusta pin).
   * Jos merkki ei lydy, palautetaan koko jono ja tyhjennetn jono.
   * <pre>
   * Kyttesimerkki: olkoon aluksi string jono,s;
   *  1)  jono = "123 456";  s = erota(jono);   => jono == "456"  s == "123"
   *  2)  jono = "123";      s = erota(jono);   => jono == ""     s == "123"
   *  3)  jono = "1 2 3";
   *  while ( jono.length() > 0 ) System.out.print(erota(jono)+",");
   *           => tulostaa 1,2,3,
   * </pre>
   * 
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param etsitakaperin  etsitnk merkki oikealta vasemmalle
   * @return               merkki edeltv jono
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono= new StringBuffer("123 456"); String s;
   *   s = erota(jono); jono.toString() === "456"; s === "123";
   * 
   *   jono = new StringBuffer($jono1); 
   *   s = erota(jono,$merkki,$suunta);
   *   jono.toString() === $jono2; s === $s;
   *   
   *   $jono1     | $merkki | $suunta | $jono2   | $s
   *   ---------------------------------------------------
   *   "123"      | ' '     | false   | ""       | "123"
   *   ---        | ' '     | false   | ""       | ""
   *   ---        | ' '     | false   | ""       | ""
   *   "123 456"  | ' '     | false   | "456"    | "123"
   *   ---        | ' '     | false   | ""       | "456"
   *   ---        | ' '     | false   | ""       | ""
   *   "12 34 6"  | ' '     | false   | "34 6"   | "12"   
   *   ---        | ' '     | false   | "6"      | "34"
   *   ---        | ' '     | false   | ""       | "6"
   *   ---        | ' '     | false   | ""       | ""
   *   "12 34 6"  | ' '     | true    | "6"      | "12 34"
   *   ---        | ' '     | true    | ""       | "6"
   *   ---        | ' '     | true    | ""       | ""
   * 
   *   erota((StringBuffer)null,' ',false) === "";
   * </pre>
   */
  public static String erota(StringBuffer jono, char merkki, boolean etsitakaperin) {
    if ( jono == null ) return "";
    int p;
    if ( etsitakaperin ) p = jono.lastIndexOf(""+merkki);
    else p = jono.indexOf(""+merkki);
    String alku;
    if ( p < 0 ) {
      alku = jono.toString();
      jono.delete(0,jono.length());
      return alku;
    }
    alku = jono.substring(0,p);
    jono.delete(0,p+1);
    return alku;
  }

  
  //****************************************************************************
  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * Alkuosa palautetaan funktion nimess ja loppuosa jtetn
   * jonoon.  Merkin etsimissuunta voidaan valita (oletuksena alusta pin).
   * Jos merkki ei lydy, palautetaan koko jono ja tyhjennetn jono.
   * <pre>
   * Kyttesimerkki: olkoon aluksi string jono,s;
   *  1)  jono = "123 456";  s = erota(jono);   => jono == "456"  s == "123"
   *  2)  jono = "123";      s = erota(jono);   => jono == ""     s == "123"
   *  3)  jono = "1 2 3";
   *  while ( jono.length() > 0 ) System.out.print(erota(jono)+",");
   *           => tulostaa 1,2,3,
   * </pre>
   * 
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param etsitakaperin  etsitnk merkki oikealta vasemmalle
   * @return               merkki edeltv jono
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono= new StringBuilder("123 456"); String s;
   *   s = erota(jono); jono.toString() === "456"; s === "123";
   * 
   *   jono = new StringBuilder($jono1); 
   *   s = erota(jono,$merkki,$suunta);
   *   jono.toString() === $jono2; s === $s;
   *   
   *   $jono1     | $merkki | $suunta | $jono2   | $s
   *   ---------------------------------------------------
   *   "123"      | ' '     | false   | ""       | "123"
   *   ---        | ' '     | false   | ""       | ""
   *   ---        | ' '     | false   | ""       | ""
   *   "123 456"  | ' '     | false   | "456"    | "123"
   *   ---        | ' '     | false   | ""       | "456"
   *   ---        | ' '     | false   | ""       | ""
   *   "12 34 6"  | ' '     | false   | "34 6"   | "12"   
   *   ---        | ' '     | false   | "6"      | "34"
   *   ---        | ' '     | false   | ""       | "6"
   *   ---        | ' '     | false   | ""       | ""
   *   "12 34 6"  | ' '     | true    | "6"      | "12 34"
   *   ---        | ' '     | true    | ""       | "6"
   *   ---        | ' '     | true    | ""       | ""
   * 
   *   erota((StringBuilder)null,' ',false) === "";
   * </pre>
   */
  public static String erota(StringBuilder jono, char merkki, boolean etsitakaperin) {
    if ( jono == null ) return "";
    int p;
    if ( etsitakaperin ) p = jono.lastIndexOf(""+merkki);
    else p = jono.indexOf(""+merkki);
    String alku;
    if ( p < 0 ) {
      alku = jono.toString();
      jono.delete(0,jono.length());
      return alku;
    }
    alku = jono.substring(0,p);
    jono.delete(0,p+1);
    return alku;
  }

  
  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @return               merkki edeltv jono
   * @example
   * <pre name="test">
   *   StringBuffer jono= new StringBuffer("123 456"); String s;
   *   s = erota(jono,' '); jono.toString() === "456"; s === "123";
   * </pre>  
   */
  public static String erota(StringBuffer jono, char merkki) {
    return erota(jono,merkki,false);
  }

  
  /** Erottaa jonosta vlilynnin kohdalta alkuosan ja loppuosan.
   * @param jono jono jota pilkotaan
   * @return     merkki edeltv jono
   */
  public static String erota(StringBuffer jono) {
    return erota(jono,' ',false);
  }

  
  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @return               merkki edeltv jono
   * @example
   * <pre name="test">
   *   StringBuilder jono= new StringBuilder("123 456"); String s;
   *   s = erota(jono,' '); jono.toString() === "456"; s === "123";
   * </pre>  
   */
  public static String erota(StringBuilder jono, char merkki) {
    return erota(jono,merkki,false);
  }

  
  /** Erottaa jonosta vlilynnin kohdalta alkuosan ja loppuosan.
   * @param jono jono jota pilkotaan
   * @return     merkki edeltv jono
   */
  public static String erota(StringBuilder jono) {
    return erota(jono,' ',false);
  }

  
  //****************************************************************************

  /**
   * Tunnetut numeromerkit
   */
  public static final String NUMEROT = "0123456789";
  
  /**
   * Kytetty desimaalisymboli
   */
  public static final char DESIMAALI = '.';
  
  /**
   * Kytetty desimaalisymboli
   */
  public static final char ETUMERKKI = '-';
  
  /**
   * Tunnetut reaaliluvun merkit
   */
  public static final String DOUBLEMERKIT = ETUMERKKI + NUMEROT + DESIMAALI;

  /**
   * Tunnetut kokonaisluvun merkit
   */
  public static final String INTMERKIT = ETUMERKKI + NUMEROT;

  //****************************************************************************

  /**
   * Erottaa merkkijonosta seuraavan reaaliluvun.  Jos lukua ei voida erottaa
   * niin palautetaan oletus
   * <pre>
   * jono = " 12.34 km"; d = erotaDouble(jono,0);  => d = 12.34  jono = " km"
   * jono = "12k34 km"; d = erotaDouble(jono,0);  => d = 12  jono = "k34 km"
   * jono = "k12.34 km"; d = erotaDouble(jono,0); => d = 0  jono = "k12.34 km"
   * </pre>
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @param numerot aluksi sallitut numerot
   * @return otettu reaaliluku
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer($jono1); erotaDouble(jono,0) ~~~ $d; jono.toString() === $jono2;
   *   
   *    $jono1    |  $d   | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 12.34 | " km" 
   *  "12k34 km"  | 12    | "k34 km"
   *  "k12.34 km" | 0     | "k12.34 km"
   *  "1.2 3.4"   | 1.2   | " 3.4"
   *   ---        | 3.4   | ""
   *  "-1.2 -3.4" | -1.2  | " -3.4"
   *   ---        | -3.4  | ""
   *  "-1-2 -3.4" | -1    | "-2 -3.4"
   *   ---        | -2    | " -3.4"
   *  ".2"        | 0.2   | ""
   *  "-.2"       | -0.2  | ""
   *  ".-2"       | 0.0   | "-2"
   *  "- 2"       | 0.0   | " 2"
   *  ""          | 0     | ""
   *
   *   jono = new StringBuffer("2112233451232"); 
   *   erotaDouble(jono,0,"123") ~~~ 2112233.0; 
   *   erotaDouble(jono,0,null) ~~~ 451232.0; 
   *   erotaDouble((StringBuffer)null,0,null) ~~~ 0.0; 
   *   
   * </pre>
   */
  public static double erotaDouble(StringBuffer jono, double oletus, String numerot) { // NOPMD onhan se monimutkainen mutta...
    if ( jono == null ) return oletus;
    String num = numerot;
    if ( num == null ) num = DOUBLEMERKIT;
    int alku = 0; int i; boolean alkuws = true;
    for (i=0; i < jono.length(); i++) {
      char c = jono.charAt(i);
      if ( alkuws && Character.isWhitespace(c) ) { alku++; continue; }
      if ( num.indexOf(c) < 0 ) break;
      if ( c == DESIMAALI ) num = NUMEROT;
      int p = num.indexOf(ETUMERKKI);
      if ( p >= 0 ) num = num.substring(0,p) + num.substring(p+1); // ei en etumerkki
      alkuws = false; // en ei saa olla ws-merkki
    }

    if ( i == 0 ) return oletus;
    num = jono.substring(alku,i);
    jono.delete(0,i);
    try {
      return Double.parseDouble(num);
    } catch (NumberFormatException ex) {
      return oletus;
    }

  }


  /**
   * Erottaa merkkijonosta seuraavan reaaliluvun.  Jos lukua ei voida erottaa
   * niin palautetaan oletus
   * <pre>
   * jono = " 12.34 km"; d = erotaDouble(jono,0);  => d = 12.34  jono = " km"
   * jono = "12k34 km"; d = erotaDouble(jono,0);  => d = 12  jono = "k34 km"
   * jono = "k12.34 km"; d = erotaDouble(jono,0); => d = 0  jono = "k12.34 km"
   * </pre>
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @param numerot aluksi sallitut numerot
   * @return otettu reaaliluku
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder($jono1); erotaDouble(jono,0) ~~~ $d; jono.toString() === $jono2;
   *   
   *    $jono1    |  $d   | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 12.34 | " km" 
   *  "12k34 km"  | 12    | "k34 km"
   *  "k12.34 km" | 0     | "k12.34 km"
   *  "1.2 3.4"   | 1.2   | " 3.4"
   *   ---        | 3.4   | ""
   *  "-1.2 -3.4" | -1.2  | " -3.4"
   *   ---        | -3.4  | ""
   *  "-1-2 -3.4" | -1    | "-2 -3.4"
   *   ---        | -2    | " -3.4"
   *  ".2"        | 0.2   | ""
   *  "-.2"       | -0.2  | ""
   *  ".-2"       | 0.0   | "-2"
   *  "- 2"       | 0.0   | " 2"
   *  ""          | 0     | ""
   *
   *   jono = new StringBuilder("2112233451232"); 
   *   erotaDouble(jono,0,"123") ~~~ 2112233.0; 
   *   erotaDouble(jono,0,null) ~~~ 451232.0; 
   *   erotaDouble((StringBuilder)null,0,"123") ~~~ 0.0; 
   *   
   * </pre>
   */
  public static double erotaDouble(StringBuilder jono, double oletus, String numerot) { // NOPMD onhan se monimutkainen mutta...
    if ( jono == null ) return oletus;
    String num = numerot;
    if ( num == null ) num = DOUBLEMERKIT;
    int alku = 0; int i; boolean alkuws = true;
    for (i=0; i < jono.length(); i++) {
      char c = jono.charAt(i);
      if ( alkuws && Character.isWhitespace(c) ) { alku++; continue; }
      if ( num.indexOf(c) < 0 ) break;
      if ( c == DESIMAALI ) num = NUMEROT;
      int p = num.indexOf(ETUMERKKI);
      if ( p >= 0 ) num = num.substring(0,p) + num.substring(p+1); // ei en etumerkki
      alkuws = false; // en ei saa olla ws-merkki
    }

    if ( i == 0 ) return oletus;
    num = jono.substring(alku,i);
    jono.delete(0,i);
    try {
      return Double.parseDouble(num);
    } catch (NumberFormatException ex) {
      return oletus;
    }

  }


  /**
   * Erottaa merkkijonosta seuraavan reaaliluvun.  Jos lukua ei voida erottaa
   * niin palautetaan oletus
   * <pre>
   * jono = " 12.34 km"; d = erotaDouble(jono,0);  => d = 12.34  jono = " km"
   * jono = "12k34 km"; d = erotaDouble(jono,0);  => d = 12  jono = "k34 km"
   * jono = "k12.34 km"; d = erotaDouble(jono,0); => d = 0  jono = "k12.34 km"
   * </pre>
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @return otettu reaaliluku
   * @see #erotaDouble(StringBuffer jono, double oletus, String numerot)
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer(" 12.34 km"); erotaDouble(jono,0) ~~~ 12.34; jono.toString() === " km"; 
   *   jono = new StringBuffer(" 12k34 km"); erotaDouble(jono,0) ~~~ 12.00; jono.toString() === "k34 km"; 
   *   jono = new StringBuffer("k12.34 km"); erotaDouble(jono,0) ~~~  0.0;  jono.toString() === "k12.34 km"; 
   * </pre>
   */
  public static double erotaDouble(StringBuffer jono, double oletus) {
    return erotaDouble(jono,oletus,DOUBLEMERKIT);
  }


  /**
   * Erottaa merkkijonosta seuraavan reaaliluvun.  Jos lukua ei voida erottaa
   * niin palautetaan oletus
   * <pre>
   * jono = " 12.34 km"; d = erotaDouble(jono,0);  => d = 12.34  jono = " km"
   * jono = "12k34 km"; d = erotaDouble(jono,0);  => d = 12  jono = "k34 km"
   * jono = "k12.34 km"; d = erotaDouble(jono,0); => d = 0  jono = "k12.34 km"
   * </pre>
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @return otettu reaaliluku
   * @see #erotaDouble(StringBuilder jono, double oletus, String numerot)
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder(" 12.34 km"); erotaDouble(jono,0) ~~~ 12.34; jono.toString() === " km"; 
   *   jono = new StringBuilder(" 12k34 km"); erotaDouble(jono,0) ~~~ 12.00; jono.toString() === "k34 km"; 
   *   jono = new StringBuilder("k12.34 km"); erotaDouble(jono,0) ~~~  0.0;  jono.toString() === "k12.34 km"; 
   * </pre>
   */
  public static double erotaDouble(StringBuilder jono, double oletus) {
    return erotaDouble(jono,oletus,DOUBLEMERKIT);
  }


  /**
   * Ottaa merkkijonosta seuraavan kokonaisluvun
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @return otettu kokonaisluku
   * @see #erotaDouble(StringBuffer jono, double oletus)
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer($jono1); erotaInt(jono,0) === $i; jono.toString() === $jono2;
   *   
   *    $jono1    |  $i   | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 12    | ".34 km" 
   *  "12k34 km"  | 12    | "k34 km"
   *  "k12.34 km" | 0     | "k12.34 km"
   *  "1.2 3.4"   | 1     | ".2 3.4"
   *   ---        | 0     | ".2 3.4"
   *  ""          | 0     | ""
   *  " 1 2 3 "   | 1     | " 2 3 "
   *  ---         | 2     | " 3 "
   *  ---         | 3     | " "
   *  ---         | 0     | ""
   *  " -1 -2 3 " | -1    | " -2 3 "
   *  ---         | -2    | " 3 "
   *  ---         | 3     | " "
   *
   *   jono = new StringBuffer("2112233451232"); 
   *   erotaInt(jono,0,"123") === 2112233; 
   * </pre>
   */
  public static int erotaInt(StringBuffer jono, int oletus) {
    return (int)Math.round(erotaDouble(jono,oletus,INTMERKIT));
  }

  
  /**
   * Ottaa merkkijonosta seuraavan kokonaisluvun
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @return otettu kokonaisluku
   * @see #erotaDouble(StringBuffer jono, double oletus)
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder($jono1); erotaInt(jono,0) === $i; jono.toString() === $jono2;
   *   
   *    $jono1    |  $i   | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 12    | ".34 km" 
   *  "12k34 km"  | 12    | "k34 km"
   *  "k12.34 km" | 0     | "k12.34 km"
   *  "1.2 3.4"   | 1     | ".2 3.4"
   *   ---        | 0     | ".2 3.4"
   *  ""          | 0     | ""
   *  " 1 2 3 "   | 1     | " 2 3 "
   *  ---         | 2     | " 3 "
   *  ---         | 3     | " "
   *  ---         | 0     | ""
   *  " -1 -2 3 " | -1    | " -2 3 "
   *  ---         | -2    | " 3 "
   *  ---         | 3     | " "
   *
   *   jono = new StringBuilder("2112233451232"); 
   *   erotaInt(jono,0,"123") === 2112233; 
   * </pre>
   */
  public static int erotaInt(StringBuilder jono, int oletus) {
    return (int)Math.round(erotaDouble(jono,oletus,INTMERKIT));
  }

  
  /**
   * Ottaa merkkijonosta seuraavan kokonaisluvun
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @param numerot numero jotka sallitaan lukuun
   * @return otettu kokonaisluku
   * @see #erotaDouble(StringBuffer jono, double oletus, String numerot)
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer("211223345"); 
   *   erotaInt(jono,0,"123") === 2112233; 
   * </pre>
   */
  public static int erotaInt(StringBuffer jono, int oletus, String numerot) {
      return (int)Math.round(erotaDouble(jono,oletus,numerot));
  }

  
  /**
   * Ottaa merkkijonosta seuraavan kokonaisluvun
   * @param jono merkkijono josta luku otetaan, jono lyhenee
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @param numerot numero jotka sallitaan lukuun
   * @return otettu kokonaisluku
   * @see #erotaDouble(StringBuilder jono, double oletus, String numerot)
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder("211223345"); 
   *   erotaInt(jono,0,"123") === 2112233; 
   * </pre>
   */
  public static int erotaInt(StringBuilder jono, int oletus, String numerot) {
      return (int)Math.round(erotaDouble(jono,oletus,numerot));
  }

  
  /**
   * Erottaa merkkijonosta seuraavan reaaliluvun.
   * @param jono merkkijono josta luku otetaan, jono ei muutu
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @return otettu reaaliluku
   * @see #erotaDouble(StringBuilder jono, double oletus)
   * @example
   * <pre name="test">
   *   erotaDouble(" 12.34 km",0) ~~~ 12.34;  
   *   erotaDouble(" 12k34 km",0) ~~~ 12.00;  
   *   erotaDouble("k12.34 km",0) ~~~  0.0;   
   *   erotaDouble((String)null,2) ~~~  2.0;   
   * </pre>
   */
  public static double erotaDouble(String jono, double oletus) {
    if ( jono == null ) return oletus;
    return erotaDouble(new StringBuilder(jono),oletus);
  }

  
  /**
   * Ottaa merkkijonosta seuraavan kokonaisluvun
   * @param jono merkkijono josta luku otetaan, jono ei muutu
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn numeroa
   * @return otettu kokonaisluku
   * @see #erotaInt(StringBuilder jono, int oletus)
   * @example
   * <pre name="test">
   *   erotaInt(" 123k",12)  === 123; 
   *   erotaInt(" k123k",12) === 12;
   *   erotaInt((String)null,2) === 2; 
   * </pre>
   */
  public static int erotaInt(String jono, int oletus) {
    if ( jono == null ) return oletus;
    return erotaInt(new StringBuilder(jono),oletus);
  }

  
  /**
   * Ottaa merkkijonosta seuraavan merkin
   * @param jono merkkijono josta merkki otetaan
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn merkki
   * @return otettu merkki
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer($jono1); erotaChar(jono,'!') === $c; jono.toString() === $jono2;
   *   
   *    $jono1    |  $c   | $jono2
   *  ----------------------------------
   *  " 12 "      | ' '   | "12 " 
   *   ---        | '1'   | "2 "
   *   ---        | '2'   | " "
   *   ---        | ' '   | ""
   *   ---        | '!'   | ""
   *    
   *  erotaChar((StringBuffer)null,'a') === 'a'  
   * </pre>
   */
  public static char erotaChar(StringBuffer jono, char oletus) {
    if ( jono == null ) return oletus;
    if ( jono.length() == 0 ) return oletus;
    char c = jono.charAt(0);
    jono.delete(0,1);
    return c;
  }

  /**
   * Ottaa merkkijonosta seuraavan merkin
   * @param jono merkkijono josta merkki otetaan
   * @param oletus arvo jota kytetn jos jonossa ei ole yhtn merkki
   * @return otettu merkki
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder($jono1); erotaChar(jono,'!') === $c; jono.toString() === $jono2;
   *   
   *    $jono1    |  $c   | $jono2
   *  ----------------------------------
   *  " 12 "      | ' '   | "12 " 
   *   ---        | '1'   | "2 "
   *   ---        | '2'   | " "
   *   ---        | ' '   | ""
   *   ---        | '!'   | ""
   *    
   *  erotaChar((StringBuilder)null,'a') === 'a'  
   * </pre>
   */
  public static char erotaChar(StringBuilder jono, char oletus) {
    if ( jono == null ) return oletus;
    if ( jono.length() == 0 ) return oletus;
    char c = jono.charAt(0);
    jono.delete(0,1);
    return c;
  }

  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * Jono trimmataan ennen palautusta (kaikki tyhjt pois)
   * Kytt oletusta jos jono on tyhj.
   * Tehty lhinn "kaveriksi" vastaaville int ja double funktiolle,
   * jotta funktiopolymorfismin avulla saadaan symmetrisi aliohjelmia
   * jolloin kutsussa on helppo vaihtaa parametrin tyyppi
   * <pre>
   *   String s = erota(jono,'|',"");
   *   double d = erota(jono,'|',0.0);
   *   s = erota(jono,'|',s);
   *   d = erota(jono,'|',d);
   * </pre>
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo jota kytetn jos palasesta tulee tyhj
   * @return               merkki edeltv jono
   * @see #erota(StringBuffer jono, char merkki, boolean etsitakaperin)
   * 
   * @example
   * <pre name="test">
   * // Testataan merkkijono erota-metodia
   *  StringBuffer jono = new StringBuffer(" Volvo    145 |  12300 e | 1 kpl");
   *  erota(jono,'|',"!") === "Volvo 145"; 
   *  erota(jono,'|',"!") === "12300 e";
   *  erota(jono,'|',"!") === "1 kpl";
   *  erota(jono,'|',"!") === "!";
   *  erota(jono,'|',null) === "";
   *
   * </pre>
   */
  public static String erota(StringBuffer jono, char merkki, String oletus) {
    String pala = erota(jono,merkki,false);
    if ( pala.length() == 0 ) pala = oletus;
    if ( pala == null ) pala = "";
    return poista_2_tyhjat(pala.trim());
  }

  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * Jono trimmataan ennen palautusta (kaikki tyhjt pois)
   * Kytt oletusta jos jono on tyhj.
   * Tehty lhinn "kaveriksi" vastaaville int ja double funktiolle,
   * jotta funktiopolymorfismin avulla saadaan symmetrisi aliohjelmia
   * jolloin kutsussa on helppo vaihtaa parametrin tyyppi
   * <pre>
   *   String s = erota(jono,'|',"");
   *   double d = erota(jono,'|',0.0);
   *   s = erota(jono,'|',s);
   *   d = erota(jono,'|',d);
   * </pre>
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo jota kytetn jos palasesta tulee tyhj
   * @return               merkki edeltv jono
   * @see #erota(StringBuilder jono, char merkki, boolean etsitakaperin)
   * 
   * @example
   * <pre name="test">
   * // Testataan merkkijono erota-metodia
   *  StringBuilder jono = new StringBuilder(" Volvo    145 |  12300 e | 1 kpl");
   *  erota(jono,'|',"!") === "Volvo 145"; 
   *  erota(jono,'|',"!") === "12300 e";
   *  erota(jono,'|',"!") === "1 kpl";
   *  erota(jono,'|',"!") === "!";
   *  erota(jono,'|',null) === "";
   *
   * </pre>
   */
  public static String erota(StringBuilder jono, char merkki, String oletus) {
    String pala = erota(jono,merkki,false);
    if ( pala.length() == 0 ) pala = oletus;
    if ( pala == null ) pala = "";
    return poista_2_tyhjat(pala.trim());
  }

  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta reaaliluvun
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo jota kytetn jos palasta ei saa reaalilukua
   * @return               merkki edeltv jono
   * @see #erota(StringBuffer jono, char merkki, String oletus)
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer("12.3,45"); erota(jono,',',9.0) ~~~ 12.3; erota(jono,',',9.0) ~~~ 45.0;
   *   jono = new StringBuffer("k,45");    erota(jono,',',9.0) ~~~ 9.0;  erota(jono,',',9.0) ~~~ 45.0;
   * </pre>
   */
  public static double erota(StringBuffer jono, char merkki, double oletus) {
    String pala = erota(jono,merkki,false);
    return erotaDouble(pala,oletus);
  }


  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta reaaliluvun
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo jota kytetn jos palasta ei saa reaalilukua
   * @return               merkki edeltv jono
   * @see #erota(StringBuilder jono, char merkki, String oletus)
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder("12.3,45"); erota(jono,',',9.0) ~~~ 12.3; erota(jono,',',9.0) ~~~ 45.0;
   *   jono = new StringBuilder("k,45");    erota(jono,',',9.0) ~~~ 9.0;  erota(jono,',',9.0) ~~~ 45.0;
   * </pre>
   */
  public static double erota(StringBuilder jono, char merkki, double oletus) {
    String pala = erota(jono,merkki,false);
    return erotaDouble(pala,oletus);
  }


  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta kokonaisluvun
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo jota kytetn jos palasta ei saa reaalilukua
   * @return               merkki edeltv jono
   * @see #erota(StringBuffer jono, char merkki, String oletus)
   *
   * @example
   * <pre name="test">
   * // Testataan polymorfista erota-metodia
   *  StringBuffer jono = new StringBuffer(" Volvo    145 |  12300 e | 1 kpl");
   *  String s=""; double d=0.0; int i=0;
   *  erota(jono,'|',s) === "Volvo 145"; 
   *  erota(jono,'|',d) ~~~ 12300.0;
   *  erota(jono,'|',i) === 1;
   *  erota(jono,'|',"a") === "a";
   * </pre>
   */
  public static int erota(StringBuffer jono, char merkki, int oletus) {
    String pala = erota(jono,merkki,false);
    return erotaInt(pala,oletus);
  }


  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta kokonaisluvun
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo jota kytetn jos palasta ei saa reaalilukua
   * @return               merkki edeltv jono
   * @see #erota(StringBuilder jono, char merkki, String oletus)
   *
   * @example
   * <pre name="test">
   * // Testataan polymorfista erota-metodia
   *  StringBuilder jono = new StringBuilder(" Volvo    145 |  12300 e | 1 kpl");
   *  String s=""; double d=0.0; int i=0;
   *  erota(jono,'|',s) === "Volvo 145"; 
   *  erota(jono,'|',d) ~~~ 12300.0;
   *  erota(jono,'|',i) === 1;
   *  erota(jono,'|',"a") === "a";
   * </pre>
   */
  public static int erota(StringBuilder jono, char merkki, int oletus) {
    String pala = erota(jono,merkki,false);
    return erotaInt(pala,oletus);
  }


  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * Jono trimmataan ennen palautusta (kaikki tyhjt pois)
   * Heitt poikkeuksen jos jono on tyhj
   * Tehty lhinn "kaveriksi" vastaaville int ja double funktiolle,
   * jotta funktiopolymorfismin avulla saadaan symmetrisi aliohjelmia
   * jolloin kutsussa on helppo vaihtaa parametrin tyyppi
   * <pre>
   *   String s = null;
   *   double d = 0.0;
   *   try {
   *     s = erotaEx(jono,'|',s);
   *     d = erotaEx(jono,'|',d);
   *   } catch ( NumberFormatException ex ) {
   *     System.out.println(ex.getMessage());
   *   }
   * </pre>
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo mikli tulos olisi tyhj
   * @return               merkki edeltv jono
   * @throws NumberFormatException jos palaa ei saada
   * @see #erota(StringBuffer jono, char merkki, boolean etsitakaperin)
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono;
   *   jono = new StringBuffer($jono1); erotaEx(jono,' ',"!") === $s; jono.toString() === $jono2
   *   
   *    $jono1    |  $s     | $jono2
   *  ----------------------------------
   *  " 12.34 km" | "!"     | "12.34 km"
   *  ---         | "12.34" | "km"
   *  ---         | "km"    | ""
   *  ---         | ""      | ---    #THROWS NumberFormatException
   *  "Kissa"     | "Kissa" | ""       
   *  ---         | ""      | ---    #THROWS NumberFormatException  
   *  
   *  erotaEx(new StringBuffer(";"),';',null) === "";
   * </pre>
   */
  public static String erotaEx(StringBuffer jono, char merkki, String oletus) throws NumberFormatException {
    if ( (jono == null) || jono.length()==0 )
      throw new NumberFormatException("jono = \"\"");
    String pala = erota(jono,merkki,false);
    if ( pala.length() == 0 ) pala = oletus;
    if ( pala == null ) pala = "";
    return poista_2_tyhjat(pala.trim());
  }

  
  /**
   * Erottaa jonosta valitun merkin kohdalta alkuosan ja loppuosan.
   * Jono trimmataan ennen palautusta (kaikki tyhjt pois)
   * Heitt poikkeuksen jos jono on tyhj
   * Tehty lhinn "kaveriksi" vastaaville int ja double funktiolle,
   * jotta funktiopolymorfismin avulla saadaan symmetrisi aliohjelmia
   * jolloin kutsussa on helppo vaihtaa parametrin tyyppi
   * <pre>
   *   String s = null;
   *   double d = 0.0;
   *   try {
   *     s = erotaEx(jono,'|',s);
   *     d = erotaEx(jono,'|',d);
   *   } catch ( NumberFormatException ex ) {
   *     System.out.println(ex.getMessage());
   *   }
   * </pre>
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         arvo mikli tulos olisi tyhj
   * @return               merkki edeltv jono
   * @throws NumberFormatException jos palaa ei saada
   * @see #erota(StringBuilder jono, char merkki, boolean etsitakaperin)
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono;
   *   jono = new StringBuilder($jono1); erotaEx(jono,' ',"!") === $s; jono.toString() === $jono2;
   *   
   *    $jono1    |  $s     | $jono2
   *  ----------------------------------
   *  " 12.34 km" | "!"     | "12.34 km"
   *  ---         | "12.34" | "km"
   *  ---         | "km"    | ""
   *  ---         | ""      | ---     #THROWS NumberFormatException
   *  "Kissa"     | "Kissa" | ""       
   *  ---         | ""      | ---     #THROWS NumberFormatException
   *    
   *  erotaEx(new StringBuilder(";"),';',null) === "";
   * </pre>
   */
  public static String erotaEx(StringBuilder jono, char merkki, String oletus) throws NumberFormatException {
    if ( (jono == null) || jono.length()==0 )
      throw new NumberFormatException("jono = \"\"");
    String pala = erota(jono,merkki,false);
    if ( pala.length() == 0 ) pala = oletus;
    if ( pala == null ) pala = "";
    return poista_2_tyhjat(pala.trim());
  }

  
  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta reaaliluvun.
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         tuo vain tiedon palautettavasta tyypist
   * @return               merkki edeltv jono
   * @throws NumberFormatException jos reaalilukua ei saada
   * @see #erotaEx(StringBuffer jono, char merkki, String oletus)
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono=null;
   *   jono = new StringBuffer($jono1); erotaEx(jono,' ',3.14) ~~~ $d; jono.toString() === $jono2;
   *   
   *    $jono1    |  $d     | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 0.0     | ---         #THROWS NumberFormatException  
   *  ---         | ---     | "12.34 km"
   *  ---         | 12.34   | "km"
   *  ---         | 0.0     | ---         #THROWS NumberFormatException
   *  ---         | ---     | ""
   *  ---         | 0.0     | ---         #THROWS NumberFormatException
   *  
   *   jono = new StringBuffer($jono1); erotaEx(jono,',',3.14) ~~~ $d; jono.toString() === $jono2;
   *   
   *    $jono1    |  $d     | $jono2
   *  ----------------------------------
   *  " 1.2, 3.4" | 1.2     | " 3.4" 
   *  ---         | 3.4     | ""
   *  ---         | 0.0     | ---         #THROWS NumberFormatException
   *  "1.2,k,3.4" | 1.2     | "k,3.4" 
   *  ---         | 0.0     | ---         #THROWS NumberFormatException
   *  ---         | ---     | "3.4"
   *  ---         | 3.4     | ""
   * </pre>
   */
  public static double erotaEx(StringBuffer jono, char merkki, double oletus) throws NumberFormatException {
    String pala = erota(jono,merkki,false);
    double d = erotaDouble(pala,Double.MIN_VALUE);
    if ( d == Double.MIN_VALUE )
      throw new NumberFormatException("jono = \"" + pala + "\""); // NOPMD - ei jaksa tehd vakiota
    return d;
  }


  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta reaaliluvun.
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         tuo vain tiedon palautettavasta tyypist
   * @return               merkki edeltv jono
   * @throws NumberFormatException jos reaalilukua ei saada
   * @see #erotaEx(StringBuilder jono, char merkki, String oletus)
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono=null;
   *   jono = new StringBuilder($jono1); erotaEx(jono,' ',3.14) ~~~ $d; jono.toString() === $jono2;
   *   
   *    $jono1    |  $d     | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 0.0     | ---        #THROWS NumberFormatException 
   *  ---         | ---     | "12.34 km"
   *  ---         | 12.34   | "km"
   *  ---         | 0.0     | ---        #THROWS NumberFormatException
   *  ---         | ---     | ""
   *  ---         | 0.0     | ---        #THROWS NumberFormatException
   *  
   *   jono = new StringBuilder($jono1); erotaEx(jono,',',3.14) ~~~ $d; jono.toString() === $jono2;
   *   
   *    $jono1    |  $d     | $jono2
   *  ----------------------------------
   *  " 1.2, 3.4" | 1.2     | " 3.4" 
   *  ---         | 3.4     | ""
   *  ---         | 0.0     | ---      #THROWS NumberFormatException
   *  "1.2,k,3.4" | 1.2     | "k,3.4" 
   *  ---         | 0.0     | ---      #THROWS NumberFormatException
   *  ---         | ---     | "3.4"
   *  ---         | 3.4     | ""
   * </pre>
   */
  public static double erotaEx(StringBuilder jono, char merkki, double oletus) throws NumberFormatException {
    String pala = erota(jono,merkki,false);
    double d = erotaDouble(pala,Double.MIN_VALUE);
    if ( d == Double.MIN_VALUE )
      throw new NumberFormatException("jono = \"" + pala + "\"");
    return d;
  }


  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta kokonaisluvun
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         tuo vain tiedon palautettavasta tyypist
   * @return               merkki edeltv jono
   * @throws NumberFormatException jos kokonaislukua ei saada
   * @see #erotaEx(StringBuffer jono, char merkki, String oletus)
   * 
   * @example
   * <pre name="test">
   *   StringBuffer jono=null;
   *   jono = new StringBuffer($jono1); erotaEx(jono,' ',314) === $i; jono.toString() === $jono2;
   *   
   *    $jono1    |  $i     | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 0       | ---    #THROWS NumberFormatException 
   *  ---         | ---     | "12.34 km"
   *  ---         | 12      | "km"
   *  ---         | 0       | ---    #THROWS NumberFormatException
   *  ---         | 0       | ---    #THROWS NumberFormatException
   *  
   *   jono = new StringBuffer($jono1); erotaEx(jono,',',314) === $i; jono.toString() === $jono2;
   *   
   *    $jono1    |  $i     | $jono2
   *  ----------------------------------
   *  " 1.2, 3.4" | 1       | " 3.4" 
   *  ---         | 3       | ""
   *  ---         | 0       | ---        #THROWS NumberFormatException
   *  "1,2,3"     | 1       | "2,3"
   *  ---         | 2       | "3"
   *  ---         | 3       | ""
   *  ---         | 0       | ---        #THROWS NumberFormatException
   *  "1,k,3"     | 1       | "k,3"      
   *  ---         | 0       | ---        #THROWS NumberFormatException 
   *  ---         | ---     | "3"
   *  ---         | 3       | ""
   *  ---         | 0       | ---        #THROWS NumberFormatException
   * </pre>
   */
  public static int erotaEx(StringBuffer jono, char merkki, int oletus) throws NumberFormatException {
    String pala = erota(jono,merkki,false);
    int i = erotaInt(pala,Integer.MIN_VALUE);
    if ( i == Integer.MIN_VALUE )
      throw new NumberFormatException("jono = \"" + pala + "\"");
    return i;
  }


  /**
   * Ottaa merkkijonosta ensin seuraavan palasen merkki-parametriin
   * saakka ja sitten ottaa tst palasta kokonaisluvun
   * @param jono           jono jota pilkotaan
   * @param merkki         merkki jonka kohdalta katkaistaan
   * @param oletus         tuo vain tiedon palautettavasta tyypist
   * @return               merkki edeltv jono
   * @throws NumberFormatException jos kokonaislukua ei saada
   * @see #erotaEx(StringBuilder jono, char merkki, String oletus)
   * 
   * @example
   * <pre name="test">
   *   StringBuilder jono=null;
   *   jono = new StringBuilder($jono1); erotaEx(jono,' ',314) === $i; jono.toString() === $jono2;
   *   
   *    $jono1    |  $i     | $jono2
   *  ----------------------------------
   *  " 12.34 km" | 0       | ---        #THROWS NumberFormatException 
   *  ---         | ---     | "12.34 km"
   *  ---         | 12      | "km"
   *  ---         | 0       | ---        #THROWS NumberFormatException
   *  ---         | 0       | ---        #THROWS NumberFormatException
   *  
   *   jono = new StringBuilder($jono1); erotaEx(jono,',',314) === $i; jono.toString() === $jono2;
   *   
   *    $jono1    |  $i     | $jono2
   *  ----------------------------------
   *  " 1.2, 3.4" | 1       | " 3.4" 
   *  ---         | 3       | ""
   *  ---         | 0       | ---        #THROWS NumberFormatException
   *  "1,2,3"     | 1       | "2,3"
   *  ---         | 2       | "3"
   *  ---         | 3       | ""
   *  ---         | 0       | ---        #THROWS NumberFormatException
   *  "1,k,3"     | 1       | "k,3"      
   *  ---         | 0       | ---        #THROWS NumberFormatException 
   *  ---         | ---     | "3"
   *  ---         | 3       | ""
   *  ---         | 0       | ---        #THROWS NumberFormatException
   * </pre>
   */
  public static int erotaEx(StringBuilder jono, char merkki, int oletus) throws NumberFormatException {
    String pala = erota(jono,merkki,false);
    int i = erotaInt(pala,Integer.MIN_VALUE);
    if ( i == Integer.MIN_VALUE )
      throw new NumberFormatException("jono = \"" + pala + "\"");
    return i;
  }



  /****************************************************************************


  private static void tulostaViiva(String s) {
    System.out.println("");
    System.out.println("=====================================================");
    System.out.println(s);
    System.out.println("=====================================================");
  }

  /**
   * Testataan Mjonot-luokkaa. 
   * @param args Ei kytss
   *
  public static void main(String[] args) { // NOPMD
      
    tulostaViiva("Testataan erota-metodia.");
    StringBuffer jono = new StringBuffer("1 2 3");
    while ( jono.length() > 0 ) {
      System.out.print(erota(jono)+",");
    }
    System.out.println("");

    tulostaViiva("Testataan erotaDouble ja erotaInt-metodeja");
    String testi[] = {" 12.34 km", "12k34", "k12.34", "12.34.56", ".34 km" };
    for ( int i = 0; i < testi.length; i++ ) {
      StringBuffer jono2 = new StringBuffer(testi[i]); // NOPMD
      double d = erotaDouble(jono2,0);
      System.out.println(fmt("\""+testi[i] + "\"",-15) + fmt(" d=" + d,-10) +
                             " jono = \"" + jono2+ "\"");
    }
    for ( int i = 0; i < testi.length; i++ ) {
      StringBuffer jono2 = new StringBuffer(testi[i]); // NOPMD
      int d = erotaInt(jono2,99);
      System.out.println(fmt("\""+testi[i] + "\"",-15) + fmt(" i=" + d,-10) +
                             " jono = \"" + jono2+ "\"");
    }

    tulostaViiva("Testataan tayta-metodia");
    for (int i=0; i<17; i++)
      System.out.println(tayta("kissa",i));

    tulostaViiva("Testataan fmt-metodia");
    for (int i=3; i<10; i++) System.out.println(fmt(1.23456,i,3));
    for (int i=3; i<7; i++) System.out.println(fmt(-1.23456,i,0));
    for (int i=9; i<14; i++) System.out.println(fmt(1234567.89,i,3));

    tulostaViiva("Testataan polymorfista erota-metodia");
    StringBuffer jono3 = new StringBuffer(" Volvo    145 |  12300 e | 1 kpl");
    String s=""; double d=0.0; int i=0;
    s = erota(jono3,'|',s);
    d = erota(jono3,'|',d);
    i = erota(jono3,'|',i);
    System.out.println("'"+s+"'  " + fmt(d,4,2) + "   " + i);

    tulostaViiva("Testataan polymorfista erotaEx-metodia");
    StringBuffer jono4 = new StringBuffer(" 23.4 | 12 | kala | 99");
    d = 0;

    for (i=0; i<5; i++ ) {
      try {
        d = erotaEx(jono4,'|',d);
      } catch (NumberFormatException ex) {
        System.out.println("Virhe: " + ex.getMessage());
      }
      System.out.println("d = " + fmt(d,4,2) + "  jono = \"" + jono4 + "\"");
    }


  }
  */
}
