package fi.jyu.mit.ohj2;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;

/**
 * Luokka tietovirtojen uudelleen suuntaamiseksi
 * @author vesal
 * 
 */
public class Suuntaaja { // NOPMD -luokkakirjasto
// #STATICIMPORT
// #import fi.jyu.mit.ohj2.*;
    
    /**
     * Rajapinta suuntajalle
     * @author vesal
     *
     */
    public interface ISuuntaaja {
        
        /** 
         * Palauttaa suuntauksen alkuperiseen tilaan. 
         */
        void palauta();
    }
    
    /**
     * Kytss olevan systeemin rivinvahdon merkkkijono
     */
    private static String NL = getNL(); // NOPMD - tarkoituksella lyhyt nimi
    
    
    /**
     * Palauttaa systeemin kytss olevan newline jonon
     * @return rivinvaihdon merkkiyhdistelm kytetyss jrjestelmss
     */
    public static String getNL() {
        if ( NL != null ) return NL;
        NL = "";
        StringOutput so = new StringOutput();
        System.out.println();
        NL = so.toString();
        so.palauta();
        return NL;
    }
    
    
    /**
     * Suuntaa inputin uudelleen ja kertoo Syotto-luokalle
     * @param is uusi tietovirta sytlle
     *  
     */
    protected static void setIn(InputStream is) {
       System.setIn(is);
       Syotto.alusta();
       Readkey.init();
    }
    
    /**
     * Luokka jolla System.in otetaan tiedostosta
     * @author vesal
     * @version 11.3.2007
     */
    public static class Input implements ISuuntaaja {
      private  static final InputStream origIn  = System.in; // NOPMD - ei ole vakio
      private InputStream stdin = null;
      
      /**
       * Asetetaan peruslukuvirta eri tiedostoon.
       * Jos nimi on null, niin sit virtaa ei suunnata uudelleen
       * @param inNimi mist tiedostosta System.in luetaan
       * @throws FileNotFoundException jos tiedostoa ei saada kyttn
       * @example
       * <pre name="test">
       * #THROWS IOException
       * #import java.io.*;
       * #import java.util.*;
       * #import static fi.jyu.mit.ohj2.VertaaTiedosto.*;
       *   kirjoitaTiedosto("hiljaa1.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\n");
       *   Input in = new Input("hiljaa1.txt");
       *   Scanner sc = new Scanner(System.in);
       *   sc.nextLine() === "33 hiljaa 1 hiipii"; 
       *   sc.nextLine() === "hyv 33 tulee";
       *   sc.hasNextLine() === false;
       *   in.palauta();
       *   tuhoaTiedosto("hiljaa1.txt"); 
       *   in = new Input(null);
       *   in.palauta();
       * </pre>
       */
      public Input(String inNimi) throws FileNotFoundException {
        // getNL();
        if ( inNimi != null ) {
          stdin = new FileInputStream(inNimi);
          setIn(stdin);
        }
      }
      
      /**
       * Palautetaan tietovirta takaisin alkuperiseen tilaan
       */
      public void palauta() {
        if ( stdin != null ) {
          try {
            stdin.close();
          } catch (IOException e) { // NOPMD
          }
          setIn(origIn);
        }
      }
      
    }

    /**
     * Luokka jolla System.out suunnataan toiseen tiedostoon
     * @author vesal
     * @version 11.3.2007
     */
    public static class Output  implements ISuuntaaja  {
      private static final PrintStream origOut = System.out; // NOPMD - ei ole vakio
      private PrintStream stdout = null;
      
      /**
       * Asetetaan perustulostusvirta eri tiedostoon.
       * Jos nimi on null, niin sit virtaa ei suunnata uudelleen
       * @param outNimi mihin System.out kirjoitetaan
       * @throws FileNotFoundException jos tiedostoa ei saada kyttn
       * @example
       * <pre name="test">
       * #THROWS IOException
       * #import java.io.*;
       * #import java.util.*;
       * #import static fi.jyu.mit.ohj2.VertaaTiedosto.*;
       *   Output out = new Output("hiljaa1.txt");
       *   System.out.println("eka");
       *   System.out.println("toka");
       *   out.palauta();
       *   vertaaFileString("hiljaa1.txt","eka\ntoka\n") === null;
       *   tuhoaTiedosto("hiljaa1.txt"); 
       *   
       *   out = new Output(null);
       *   out.palauta();
       * </pre>
       */
      public Output(String outNimi) throws FileNotFoundException {
        // getNL();
        if ( outNimi != null ) {
          stdout = new PrintStream(new FileOutputStream(outNimi));
          System.setOut(stdout);
        }
      }
      
      /**
       * Palautetaan tietovirta takaisin alkuperiseen tilaan
       */
      public void palauta() {
        if (stdout != null ) {
          stdout.close();
          System.setOut(origOut);
        }
      }
      
    }
    
    /**
     * Luokka sytn lukemiseksi merkkijonosta
     * @author vesal
     * @version 2.2.2008
     * @example
     * <pre name="test">
     *  StringInput si = new StringInput("kissa\nkoira");
     *  StringOutput so = new StringOutput();
     *  Syotto.kysy("Mik") === "kissa";
     *  Syotto.kysy("Mik") === "koira";
     *  Syotto.kysy("Mik") === "";
     *  si = new StringInput("12\n13");
     *  Syotto.kysy("Luku",0) === 12;
     *  Syotto.kysy("Luku",0) === 13;
     *  Syotto.kysy("Luku",0) === 0;
     *  si.palauta();
     *  so.palauta();
     * </pre>
     *
     */
    public static class StringInput implements ISuuntaaja {
        private static final InputStream origIn  = System.in; // NOPMD ei ole vakio
        private ByteArrayInputStream byteinput;
        

        /**
         * Alustetataan lukutietovirta 
         * @param inputString merkkijonojosta input otetaan
         */
        public StringInput(String inputString) {
            byteinput = new ByteArrayInputStream(inputString.getBytes());
            setIn(byteinput);
        }

        /**
         * Palautetaan tietovirta takaisin alkuperiseen tilaan
         */
        public void palauta() {
            setIn(origIn);
        }
        
        /**
         * Laitetaan syttn uusi merkkijono jota luetaan-
         * @param inputString  merkkijonojosta input otetaan
         * @example
         * <pre name="test">
         *  StringInput si = new StringInput("kissa\nkoira");
         *  StringOutput so = new StringOutput();
         *  Syotto.kysy("Mik") === "kissa";
         *  Syotto.kysy("Mik") === "koira";
         *  Syotto.kysy("Mik") === "";
         *  si.input("12\n13");
         *  Syotto.kysy("Luku",0) === 12;
         *  Syotto.kysy("Luku",0) === 13;
         *  Syotto.kysy("Luku",0) === 0;
         *  si.palauta();
         *  so.palauta();
         * </pre>
             */
        public void input(String inputString) {
            byteinput = new ByteArrayInputStream(inputString.getBytes());
            setIn(byteinput);
        }
    }
    
    
    /**
     * Luokka tulostuksen siirtmiseksi merkkijonoon
     * @author vesal
     * @version 2.2.2008
     *
     */
    public static class StringOutput implements ISuuntaaja {
        private static final PrintStream origOut = System.out; // NOPMD - ei ole vakio
        private final ByteArrayOutputStream byteoutput;
        
        /**
         * Alutetataan kirjoitustietovirta 
         */
        public StringOutput() {
            // if ( NL != null ) getNL();
            byteoutput = new ByteArrayOutputStream();
            PrintStream ps = new PrintStream(byteoutput); 
            System.setOut(ps);
        }

        /**
         * Palautetaan tietovirta takaisin alkuperiseen tilaan
         */
        public void palauta() {
            System.setOut(origOut);
        }
        
        /**
         * Palautetaan toistaiseksi tulostettu tieto merkkijonona
         * @return tulostettu tieto
         * @example
         * <pre name="test">
         *   String NL = getNL();
         *   StringOutput so = new StringOutput();
         *   System.out.println("eka"); 
         *   System.out.println("toka");
         *   so.toString() === "eka"+NL+"toka"+NL; 
         *   System.out.println("kolmas");
         *   so.toStringReset() === "eka"+NL+"toka"+NL+"kolmas"+NL;
         *   so.toString() === "";
         *   System.out.println("neljs");
         *   so.toStringReset() === "neljs"+NL;
         *   System.out.print("viides\nkuudes");
         *   so.toStringReset() === "viides\nkuudes";
         *   System.out.printf("viides%nkuudes");
         *   so.toStringReset() === "viides"+NL+"kuudes";
         *   so.palauta();
         * </pre>
         */
        @Override
        public String toString() {
            return byteoutput.toString();
        }
        
        /**
         * Palautetaan toistaiseksi tulostettu tieto merkkijonona
         * ja tyhjennetn tietovirta
         * @return tulostettu tieto
         */
        public String toStringReset() {
            String result = byteoutput.toString();
            reset();
            return result;
        }
        
        /**
         * Tyhjent toistaiseksi tulostetun osan
         */
        public void reset() {
            byteoutput.reset();
        }
        
        /**
         * Kirjoittaa sislln tietovirtaan 
         * @param out virta johon kirjoitetaan
         * @throws IOException jos joku menee pieleen
         * @example
         * <pre name="test">
         * #THROWS IOException
         * PrintStream fs = Tiedosto.avaa_kirjoittamista_varten_stream("hiljaa1.txt");
         * StringOutput so = new StringOutput();
         * System.out.println("eka"); 
         * System.out.println("toka");
         * so.writeTo(fs);
         * fs.close();
         * so.palauta();
         * vertaaFileString("hiljaa1.txt","eka\ntoka\n") === null;
         * tuhoaTiedosto("hiljaa1.txt"); 
         * </pre>
         */
        public void writeTo(OutputStream out) throws IOException {
           byteoutput.writeTo(out); 
        }
        
        /**
         * Palauttaa alkuperisen tietovirran
         * @return alkuperinen tietovirta
         */
        public PrintStream getOrigOut() {
            return origOut;
        }
        
        /**
         * Vertaa tuloksen sislt jonoon ja palauttaa eron
         * tai null jos samat.  Tyhjent tulosteen.
         * @param verrattava jono johon output-jonon sislt verrataan
         * @return null jos samat, muuten 1. ero
         * @example
         * <pre name="test">
         *   StringOutput so = new StringOutput();
         *   System.out.println("eka"); 
         *   System.out.println("toka");
         *   so.ero("eka\ntoka\n")         === null; 
         *   System.out.println("kolmas");
         *   so.ero("eka\ntoka\nkolmas\n") === "Ero riveiss 1: kolmas ja eka";
         *   so.toString()                 === "";
         *   System.out.println("neljs");
         *   so.ero("neljs\n")            === null;
         *   System.out.print("viides\nkuudes");
         *   so.ero("viides\nkuudes")      === null;
         *   System.out.printf("viides%nkuudes");
         *   so.ero("viides\nkuudes")      === null;
         *   so.palauta();
         *   so.getOrigOut() == System.out === true;
         * </pre>
         */
        public String ero(String verrattava) {
            return VertaaTiedosto.vertaaString(toStringReset(), verrattava);
        }
        
    }
    
    
    
    /**
     * Luokka jolla System.in ja System.out suunnataan toiseen tiedostoon
     * @author vesal
     * @version 11.3.2007
     */
    public static class InOut implements ISuuntaaja {
      private final Input in;
      private final Output out;
      
      /**
       * Asetetaan perusluku- ja tulostusvirta eri tiedostoon.
       * Jos jompikumpi nimi on null, niin sit virtaa ei suunnata uudelleen
       * @param inNimi mist tiedostosta System.in luetaan
       * @param outNimi mihin System.out kirjoitetaan
       * @throws FileNotFoundException jos tiedostoa ei saada kyttn
       * @example
       * <pre name="test">
       * #THROWS IOException
       *   kirjoitaTiedosto("hiljaa1.txt", "eka\ntoka\n");
       *   InOut io = new InOut("hiljaa1.txt","hiljaa2.txt");
       *   Syotto.kysy("1.") === "eka";
       *   Syotto.kysy("2.") === "toka";
       *   Syotto.kysy("3.") === "";
       *   io.palauta();
       *   tuhoaTiedosto("hiljaa1.txt"); 
       *   vertaaFileString("hiljaa2.txt","1. >2. >3. >") === null;
       *   tuhoaTiedosto("hiljaa2.txt"); 
       * </pre>
       */
      public InOut(String inNimi, String outNimi) throws FileNotFoundException {
          in = new Input(inNimi);
          out = new Output(outNimi);
      }
      
      /**
       * Palautetaan tietovirrat takaisin alkuperiseen tilaan
       */
      public void palauta() {
          in.palauta();
          out.palauta();
      }
      
    }
    
    /** 
     * Testataan suuntaamista
     * @param args ei kytss
     */
/*    
    public static void main(String[] args) {
        StringInput si = new StringInput("kissa\nkoira\nkana");
        StringOutput so = new StringOutput();
        String s1 = Syotto.kysy("Mik");
        String so1 = so.toString();
        so.getOrigOut().println("so1=" + so1);
        si = new StringInput("12\n13\n15\n16\n17");
        String s2 = Syotto.kysy("Kuka");
        //si.palauta();
        String s3 = Syotto.kysy("Mik");
        so.reset();
        so.getOrigOut().println(s1 + "|" + s2 + "|" + s3);
        String s4 = Syotto.kysy("Kuis");
        System.out.println(s4);
        si = new StringInput("12\n13\n15\n16\n17");
        Scanner sc = new Scanner(System.in);
        String s5 = sc.nextLine();
        String s6 = Syotto.kysy("No");
        si.palauta();
        String so2 = so.toString();
        so.palauta();
        System.out.println(so1 + "|" + so2 + "|" + s5 + "|" + s6);
    }
*/    
}
