package fi.jyu.mit.ohj2;
import java.io.*;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.net.MalformedURLException;
import java.net.URL;

/**
 * <p>Title: Tiedosto</p>
 * <p>Description: Aliohjelmia tiedostojen ksittelyyn</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: jyu </p>
 * @author Vesa Lappalainen
 * @version 1.0, 25.02.2003
 * @version 1.1, 09.03.2002
 * @version 1.2  12.08.2009 - netist luku
 * @version 1.3  06.11.2010 - lue, joka lukee netista tai tiedostosta
 */

public class Tiedosto {


  /**
   * Avaa tiedoston lukemista varten
   * @param nimi avattavan tiedoston nimi
   * @return avattu puskuroitu tiedostolukija tai null jos ei aukea
   */
  public static BufferedReader avaa_lukemista_varten(String nimi) { // NOPMD alleviiva ok hist. syist
    try {
      return new BufferedReader(new FileReader(nimi));
    } catch (FileNotFoundException ex) {
      return null;
    }
  }


  /**
   * Avaa tiedoston kirjoittamista varten.
   * @param nimi   avattavan tiedoston nimi
   * @param jatka  jatketaanko edellisen tiedoston pern (true) vai ei (false)
   * @return       avattu PrintWriter-olio tai null jos ei aukea
   */
  public static PrintWriter avaa_kirjoittamista_varten(String nimi,boolean jatka) {  // NOPMD alleviiva ok hist. syist
    PrintWriter f = null;
      try {
        f = new PrintWriter(new FileWriter(nimi,jatka));
      } catch (IOException ex) {
        return null;
      }
    return f;
  }

  /**
   * Avaa tiedoston kirjoittamista varten.
   * @param nimi   avattavan tiedoston nimi
   * @return       avattu PrintWriter-olio tai null jos ei aukea
   */
  public static PrintWriter avaa_kirjoittamista_varten(String nimi) {  // NOPMD alleviiva ok hist. syist
    return avaa_kirjoittamista_varten(nimi,false);
  }


  /**
   * Avaa tiedoston kirjoittamista varten.
   * @param nimi   avattavan tiedoston nimi
   * @param jatka  jatketaanko edellisen tiedoston pern (true) vai ei (false)
   * @return       avattu PrintStream-olio tai null jos ei aukea
   */
  public static PrintStream avaa_kirjoittamista_varten_stream(String nimi,boolean jatka) {  // NOPMD alleviiva ok hist. syist
    PrintStream f = null;
    try {
      f = new PrintStream(new FileOutputStream(nimi,jatka));
    }  catch (FileNotFoundException ex) {
      return null;
    }
    return f;
  }

  /**
   * Avaa tiedoston kirjoittamista varten.
   * @param nimi   avattavan tiedoston nimi
   * @return       avattu PrintStream-olio tai null jos ei aukea
   */
  public static PrintStream avaa_kirjoittamista_varten_stream(String nimi) {  // NOPMD alleviiva ok hist. syist
    return avaa_kirjoittamista_varten_stream(nimi,false);
  }


  /**
   * Luo tarvittaessa uuden PrintStream-virran OutputStream virrasta.
   * Jos os on valmiiksi PrintStream, niin tehdn vain tyypinmuunnos
   * @param os virta josta luodaan PrintStream virta
   * @return os muutettuna PrintStrean-virraksi
   */
  public static PrintStream getPrintStream(OutputStream os) {  // NOPMD alleviiva ok hist. syist
    return (os instanceof PrintStream)?(PrintStream)os : new PrintStream(os) ;
  }


  /**
   * Tulostetaan tietovirtaan rivinvaihdolla erotettuna kaikki listan rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   */
  public static void println(PrintStream out, List<String> rivit) {
      for (String s : rivit)
          out.println(s);
  }
  
  
  /**
   * Tulostetaan tietovirtaan rivinvaihdolla erotettuna kaikki taulukon rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   */
  public static void println(PrintStream out, String[] rivit) {
      for (String s : rivit)
          out.println(s);
  }
  
  
  /**
   * Tulostetaan tietovirtaan erottimella erotettuna kaikki listan rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   * @param erotin jolla rivit erotetaan
   */
  public static void print(PrintStream out, List<String> rivit, String erotin) {
      String vali = "";
      for (String s : rivit) {
          out.println(vali+s);
          vali = erotin;
      }    
  }
  
  
  /**
   * Tulostetaan tietovirtaan erottimella erotettuna kaikki taulukon rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   * @param erotin jolla rivit erotetaan
   */
  public static void print(PrintWriter out, String[] rivit, String erotin) {
      String vali = "";
      for (String s : rivit) {
          out.println(vali+s);
          vali = erotin;
      }    
  }
  
  
  /**
   * Tulostetaan tietovirtaan rivinvaihdolla erotettuna kaikki listan rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   */
  public static void println(PrintWriter out, List<String> rivit) {
      for (String s : rivit)
          out.println(s);
  }
  
  
  /**
   * Tulostetaan tietovirtaan rivinvaihdolla erotettuna kaikki taulukon rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   */
  public static void println(PrintWriter out, String[] rivit) {
      for (String s : rivit)
          out.println(s);
  }
  
  
  /**
   * Tulostetaan tietovirtaan erottimella erotettuna kaikki listan rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   * @param erotin jolla rivit erotetaan
   */
  public static void print(PrintWriter out, List<String> rivit, String erotin) {
      String vali = "";
      for (String s : rivit) {
          out.println(vali+s);
          vali = erotin;
      }    
  }
  
  
  /**
   * Tulostetaan tietovirtaan erottimella erotettuna kaikki taulukon rivit
   * @param out tietovirta johon tulostetaan
   * @param rivit tulostettavat rivit
   * @param erotin jolla rivit erotetaan
   */
  public static void print(PrintStream out, String[] rivit, String erotin) {
      String vali = "";
      for (String s : rivit) {
          out.println(vali+s);
          vali = erotin;
      }    
  }
  
  
  /**
   * Funktiolla luetaan tekstitiedosto ja palautetaan se merkkijonolistana
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param f avattu luettava tietovirta
   * @param rivit lista johon tiedoston rivit listn. 
   * @return null jos tiedostoa ei saa luettua, muuten tiedoston sislt
   */
  public static Collection<String> lueTiedosto(BufferedReader f, Collection<String> rivit) {
    if ( f == null ) return null;
    
    String rivi;
    
    try {
        while ( ( rivi = f.readLine() ) != null ) {
            rivit.add(rivi);            
        }
    } catch (IOException e) {
        return null;
    } finally {
        try {
            f.close();
        } catch (IOException e) {
            return null;  // NOPMD
        }        
    }
    return rivit;
  }
  
  
  /**
   * Funktiolla luetaan tekstitiedosto ja palautetaan se merkkijonolistana
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param nimi luettavan tiedoston nimi
   * @param rivit lista johon tiedoston rivit listn. 
   * @return null jos tiedostoa ei saa luettua, muuten tiedoston sislt
   */
  public static Collection<String> lueTiedosto(String nimi, Collection<String> rivit) {
      BufferedReader f = avaa_lukemista_varten(nimi);
      return lueTiedosto(f, rivit);
  }
  
  
  /**
   * Funktiolla luetaan tekstitiedosto ja palautetaan se merkkijonolistana
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param nimi luettavan tiedoston nimi
   * @return null jos tiedostoa ei saa luettua, muuten tiedoston sislt
   */
  public static List<String> lueTiedostoListaan(String nimi) {
    ArrayList<String> rivit = new ArrayList<String>();
    return (List<String>)lueTiedosto(nimi, rivit);
  }
  

  /**
   * Funktiolla muutetaan tietorakenne merkkijonotaulukoksi
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param rivit muutettava tietorakenne
   * @return null jos rakenne on null, muuten rakenteen sislt
   */
  public static String[] toArray(Collection<String> rivit) {
	  if ( rivit == null ) return null;
	  String [] tulosrivit = new String[rivit.size()];
	  int i=0;
	  for (String rivi:rivit) tulosrivit[i++] = rivi;
	  return tulosrivit;
  }
  
  
  /**
   * Funktiolla luetaan tekstitiedosto ja palautetaan se merkkijonotaulukkona.
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param nimi luettavan tiedoston nimi
   * @return null jos tiedostoa ei saa luettua, muuten tiedoston sislt
   */
  public static String[] lueTiedosto(String nimi) {
      List<String> rivit = lueTiedostoListaan(nimi);
      return toArray(rivit);
  }
  

  /**
   * Funktiolla luetaan nettisivun sislt ja palautetaan se merkkijonolistana
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param url luettavan tiedoston nimi
   * @param rivit lista johon sivun rivit listn. 
   * @return null jos sivua ei saa luettua, muuten sivun sislt
   */
  public static Collection<String> lueNetista(String url, Collection<String> rivit) {
      try {
		URL sivu = new URL(url);
		  InputStream in = sivu.openStream();
		  Reader reader = new InputStreamReader(in);
		  BufferedReader f = new BufferedReader(reader);
		  return lueTiedosto(f,rivit);
	} catch (MalformedURLException e) {
		return null;
	} catch (IOException e) {
		// TODO Auto-generated catch block
		return null;
	}
  }
  
  
  /**
   * Funktiolla luetaan nettisivun sislt ja palautetaan se merkkijonolistana
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param url luettavan tiedoston nimi
   * @return null jos sivua ei saa luettua, muuten sivun sislt
   */
  public static List<String> lueNetistaListaan(String url) {
	  List<String> rivit = new ArrayList<String>();
	  return (List<String>)lueNetista(url, rivit);
  }
  
  
  /**
   * Funktiolla luetaan nettisivun sislt ja palautetaan se merkkijonotaulukkona
   * Ei saa kytt isojen tiedostojen ksittelyyn! 
   * @param url luettavan tiedoston nimi
   * @return null jos sivua ei saa luettua, muuten sivun sislt
   */
  public static String[] lueNetista(String url) {
      List<String> rivit = lueNetistaListaan(url);
      return toArray(rivit);
  }

  
  /**
   * Palauttaa onko nimi URL vai tavallinen tiedoston nimi
   * @param nimi tutkittava nimi
   * @return true jos URL, muuten false
   * @example
   * <pre name="test">
   *   onkoURL("kissa.txt") === false;
   *   onkoURL("http://i.jyu.fi") === true;
   *   onkoURL("https://i.jyu.fi") === true;
   *   onkoURL("httpko.txt") === false;
   * </pre>
   */
  public static boolean onkoURL(String nimi) {
      if ( nimi.startsWith("http:") ) return true;
      if ( nimi.startsWith("https:") ) return true;
      return false;
  }
  
  
  /**
   * Lukee tiedosta tai netist jos nimi alkaa http
   * @param nimi tiedoston nimi tai URL
   * @return taulukko luetuista merkkijonoista
   */
  public static String[] lue(String nimi) {
     if ( onkoURL(nimi) )  return lueNetista(nimi);
     return lueTiedosto(nimi);
  }
  
  
  /**
   * Lukee tiedosta tai netist jos nimi alkaa http
   * @param nimi tiedoston nimi tai URL
   * @return lista luetuista merkkijonoista
   */
  public static List<String> lueListaan(String nimi) {
     if ( onkoURL(nimi) )  return lueNetistaListaan(nimi);
     return lueTiedostoListaan(nimi);
  }
  
  
  /**
   * Kirjoitetaan merkkijonotaulukko tiedostoon.
   * @param nimi tiedoston nimi
   * @param rivit kirjoitettavat rivit
   * @param jatka jatketaanko vanhan tiedoston pern (true) vai ei (false) 
   * @return false jos ei onnistu ja true jos onnistuu
   */
  public static boolean kirjoitaTiedosto(String nimi, String rivit[], boolean jatka) {
      PrintWriter f = avaa_kirjoittamista_varten(nimi, jatka);
      if ( f == null ) return false;
      try {
          for (String rivi : rivit)
              f.println(rivi);
      } finally {  
          f.close();
      }  
      return true;   
  }
  
  
  /**
   * Kirjoitetaan merkkijonotaulukko tiedostoon.
   * @param nimi tiedoston nimi
   * @param rivit kirjoitettavat rivit
   * @return false jos ei onnistu ja true jos onnistuu
   */
  public static boolean kirjoitaTiedosto(String nimi, String rivit[]) {
      return kirjoitaTiedosto(nimi, rivit, false);
  }
  
  
  /**
   * Kirjoitetaan merkkijonolista tiedostoon.
   * @param nimi tiedoston nimi
   * @param rivit kirjoitettavat rivit
   * @param jatka jatketaanko vanhan tiedoston pern (true) vai ei (false) 
   * @return false jos ei onnistu ja true jos onnistuu
   */
  public static boolean kirjoitaTiedosto(String nimi, Iterable<String> rivit, boolean jatka) {
      PrintWriter f = avaa_kirjoittamista_varten(nimi, jatka);
      if ( f == null ) return false;
      try {
          for (String rivi : rivit)
              f.println(rivi);
      } finally {  
          f.close();
      }  
      return true;   
  }
  
  
  /**
   * Kirjoitetaan merkkijonolista tiedostoon.
   * @param nimi tiedoston nimi
   * @param rivit kirjoitettavat rivit
   * @return false jos ei onnistu ja true jos onnistuu
   */
  public static boolean kirjoitaTiedosto(String nimi, Iterable<String> rivit) {
      return kirjoitaTiedosto(nimi, rivit, false);
  }
  
}
