package fi.jyu.mit.ohj2;
import java.io.*;
import java.util.Scanner;

/**
 * Aliohjelma kahden tekstitiedoston vertaamiseksi.
 * Kytt lhinn JUnit testeiss esim. seuraavasti:
 * <pre>
 *     VertaaTiedosto.kirjoitaTiedosto("hiljaa.txt", 
 *         "33 hiljaa 1 hiipii\n"+
 *         "hyv 33 tulee\n"+
 *         "36 1 3 5 55\n"+
 *         "nyt 33 riitt\n");    
 *     VertaaTiedosto.kirjoitaTiedosto("hiljaayli30.txt", 
 *         "33 hiljaa 1 hiipii\n"+
 *         "36 1 3 5 55\n");
 *     VertaaTiedosto.tuhoaTiedosto("tulos.txt");
 *     TulYli30.main(new String[]{"hiljaa.txt","tulos.txt"});
 *     String virhe=VertaaTiedosto.vertaa("tulos.txt","hiljaayli30.txt");
 *     if ( virhe != null ) fail(virhe);
 * </pre>
 * @author vesal
 * @version 10.3.2007
 */
public class VertaaTiedosto { // NOPMD Cyclomatic
  /**
   * Verrataan kahta tekstitiedostoa ja heti kun tulee ensimmin poikkeava 
   * rivi palautetaanvirhe.  Lopussa olevat pelkk yksi rivi ei tee eroa.
   * @param nimi1 1. verrattavan tiedoston nimi
   * @param nimi2 2. verrattavan tiedoston nimi
   * @return ensimminen eroavaisuus joka lytyy. Null jos ei eroja.
   * @throws IOException jos lukemisessa tapahtuu virhe.
   * @example
   * <pre name="test">
   * #THROWS IOException
   * #import java.io.*;
   * #STATICIMPORT
   *     kirjoitaTiedosto("hiljaa1.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\n");
   *         
   *     kirjoitaTiedosto("hiljaa2.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\n"); 
   *     vertaaFileFile("hiljaa1.txt","hiljaa2.txt") === null;
   *
   *     kirjoitaTiedosto("hiljaa2.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee"); 
   *     vertaaFileFile("hiljaa1.txt","hiljaa2.txt") === null;
   *
   *     kirjoitaTiedosto("hiljaa2.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\n\n\n");
   *     vertaaFileFile("hiljaa1.txt","hiljaa2.txt") === "Rivi 3: hiljaa1.txt loppui ensin, hiljaa2.txt on ";
   *
   *     kirjoitaTiedosto("hiljaa2.txt", "33 hiljaa 1 hiipii\nhyv 34 tulee\n");
   *     vertaaFileFile("hiljaa1.txt","hiljaa2.txt") === "Ero riveiss 2: hyv 33 tulee ja hyv 34 tulee";
   *     
   *     kirjoitaTiedosto("hiljaa2.txt", "33 hiljaa 1 hiipii\n");
   *     vertaaFileFile("hiljaa1.txt","hiljaa2.txt") === "Rivi 2: hiljaa2.txt loppui ensin, hiljaa1.txt on hyv 33 tulee";
   *
   *     kirjoitaTiedosto("hiljaa2.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\nja 34 tulee\n");
   *     vertaaFileFile("hiljaa1.txt","hiljaa2.txt") === "Rivi 3: hiljaa1.txt loppui ensin, hiljaa2.txt on ja 34 tulee";
   *     vertaaFileFile("hiljaa1.txt","hiljaa3.txt") === "Tiedosto ei aukea: hiljaa3.txt";
   *     vertaaFileFile("hiljaa4.txt","hiljaa2.txt") === "Tiedosto ei aukea: hiljaa4.txt";
   *     
   *     tuhoaTiedosto("hiljaa1.txt");
   *     tuhoaTiedosto("hiljaa2.txt");
   * 
   * </pre>
   */
  public static String vertaaFileFile(String nimi1, String nimi2) throws IOException { // NOPMD
    BufferedReader f1 = null;
    BufferedReader f2 = null;
    try {
      f1 = Tiedosto.avaa_lukemista_varten(nimi1);
      f2 = Tiedosto.avaa_lukemista_varten(nimi2);
      if ( f1 == null ) return "Tiedosto ei aukea: " + nimi1;
      if ( f2 == null ) return "Tiedosto ei aukea: " + nimi2;
      int n = 1;

      while (true) {
        String s1 = f1.readLine();
        String s2 = f2.readLine();
        if (s1 == null && s2 == null ) return null;
        if (s1 != null && s2 == null ) return "Rivi "+ n + ": " + nimi2 + " loppui ensin, " + nimi1 + " on "+ s1; // NOPMD
        if (s1 == null && s2 != null)  return "Rivi "+ n + ": " + nimi1 + " loppui ensin, " + nimi2 + " on "+ s2; // NOPMD
        if ( s1.compareTo(s2) != 0 )   return "Ero riveiss " + n + ": " + s1 + " ja " + s2; // NOPMD
        n++;
      }
    } finally {
      if (f1 != null)
        f1.close();
      if (f2 != null)
        f2.close();
    }
  }

  /**
   * Verrataan tekstitiedostoa merkkijonoon ja heti kun tulee ensimminen poikkeava 
   * rivi palautetaan virhe.  Lopussa olevat pelkk yksi tyhk rivi eo tee eroa.
   * @param nimi1 1. verrattavan tiedoston nimi
   * @param ss2 2. verrattava sislt
   * @return ensimminen eroavaisuus joka lytyy. Null jos ei eroja.
   * @throws IOException jos lukemisessa tapahtuu virhe.
   * <pre name="test">
   * #THROWS IOException
   * #import java.io.*;
   * #STATICIMPORT
   *     kirjoitaTiedosto("hiljaa1.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\n");
   *         
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\nhyv 33 tulee\n") === null;
   *
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\nhyv 33 tulee") === null;
   *
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\nhyv 33 tulee\n\n") === "Rivi 3: hiljaa1.txt loppui ensin, jono on ";
   *
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\nhyv 33 tulee\n\n\n") === "Rivi 3: hiljaa1.txt loppui ensin, jono on ";
   *
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\nhyv 34 tulee\n") === "Ero riveiss 2: hyv 33 tulee ja hyv 34 tulee";
   *     
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\n") === "Rivi 2: Jono loppui ensin, hiljaa1.txt on hyv 33 tulee";
   *
   *     vertaaFileString("hiljaa1.txt","33 hiljaa 1 hiipii\nhyv 33 tulee\nja 34 tulee\n") === "Rivi 3: hiljaa1.txt loppui ensin, jono on ja 34 tulee";
   *     vertaaFileString("hiljaa1.txt",null) === "Jono on null"; 
   *     vertaaFileString("hiljaa4.txt","hiljaa2.txt") === "Tiedosto ei aukea: hiljaa4.txt";
   *     
   *     tuhoaTiedosto("hiljaa1.txt");
   * 
   * </pre>
   */
  public static String vertaaFileString(String nimi1, String ss2) throws IOException { // NOPMD
    if ( ss2 == null ) return "Jono on null";
    BufferedReader f1 = null;
    Scanner f2 = new Scanner(ss2);
    int n = 1;
    try {
      f1 = Tiedosto.avaa_lukemista_varten(nimi1);
      if ( f1 == null ) return "Tiedosto ei aukea: " + nimi1;
      
      while (true) {
        String s1 = f1.readLine();
        boolean b2 = f2.hasNextLine();
        String s2 = null;
        if ( b2 ) s2 = f2.nextLine();
        if (s1 == null && !b2)         return null;
        if (s1 != null && !b2 )        return "Rivi "+ n + ": " + "Jono loppui ensin, " + nimi1 + " on "+ s1;
        if (s1 == null &&  b2)         return "Rivi "+ n + ": " + nimi1 + " loppui ensin, "  + "jono on "+ s2;
        if ( s1.compareTo(s2) != 0 )   return "Ero riveiss " + n + ": " + s1 + " ja " + s2;
        n++;
      }
    } finally {
      if (f1 != null)
        f1.close();
    }
  }

  /**
   * Verrataan kahta tekstitiedoston kaltaista merkkijonoa ja heti kun tulee ensimminen poikkeava 
   * rivi palautetaan virhe.  Lopussa oleva yksi tyhj rivi ei tee eroa.
   * @param ss1 1. verrattava sislt
   * @param ss2 2. verrattava sislt
   * @return ensimminen eroavaisuus joka lytyy. Null jos ei eroja.
   * <pre name="test">
   * #THROWS IOException
   * #import java.io.*;
   * #STATICIMPORT
   *     vertaaString2("","\n") === "Rivi 1: 1. loppui ensin, 2. on ";
   *     vertaaString2("kissa\n","kissa") === null;
   *     vertaaString2("\n\n","\n") === "Rivi 2: 2. loppui ensin, 1. on ";
   *     vertaaString2("\r\n","\n") === null;
   *     vertaaString2("\r\n","\n\n") === "Rivi 2: 1. loppui ensin, 2. on ";
   *     vertaaString2("33 hiljaa 1 hiipii\r\nhyv 33 tulee\r\n","33 hiljaa 1 hiipii\nhyv 33 tulee\n") === null;
   *     vertaaString2("a b","a c") === "Ero riveiss 1: a b ja a c"
   *     vertaaString2(null,null) === null;
   *     vertaaString2(null," ") === "1. on null";
   *     vertaaString2(" ",null) === "2. on null";
   * </pre>
   */
  public static String vertaaString2(String ss1, String ss2)  { // NOPMD
    if ( ss1 == null & ss2 == null ) return null;
    if ( ss1 == null ) return "1. on null";
    if ( ss2 == null ) return "2. on null";
    Scanner f1 = new Scanner(ss1);
    Scanner f2 = new Scanner(ss2);
    int n = 1;
    while (true) {
      String s1 = null;
      String s2 = null;
      boolean b1 = f1.hasNextLine(); 
      boolean b2 = f2.hasNextLine(); 
      if ( b1 ) s1 = f1.nextLine();
      if ( b2 ) s2 = f2.nextLine();
      if ( !b1 && !b2)             return null;
      if (  b1 && !b2 )            return "Rivi "+ n + ": " + "2. loppui ensin, 1. on "+ s1;
      if ( !b1 &&  b2)             return "Rivi "+ n + ": " + "1. loppui ensin, 2. on "+ s2;
      if ( s1.compareTo(s2) != 0 ) return "Ero riveiss " + n + ": " + s1 + " ja " + s2;
      n++;
    }
  }

  /**
   * Verrataan kahta tekstitiedoston kaltaista merkkijonoa ja heti kun tulee ensimminen poikkeava 
   * rivi palautetaan virhe.  Lopussa olevat pelkt tyhjt rivit merkitsevt.
   * @param ss1 1. verrattava sislt
   * @param ss2 2. verrattava sislt
   * @return ensimminen eroavaisuus joka lytyy. Null jos ei eroja.
   * <pre name="test">
   * #THROWS IOException
   * #import java.io.*;
   * #STATICIMPORT
   *     vertaaString("kissa\n","kissa") === "Rivi 2: 2. loppui ensin, 1. on ";
   *     vertaaString("","\n") === "Rivi 1: 1. loppui ensin, 2. on ";
   *     vertaaString("\n\n","\n") === "Rivi 3: 2. loppui ensin, 1. on ";
   *     vertaaString("\r\n","\n") === null;
   *     vertaaString("\r","\n")   === null;
   *     vertaaString("\r\n","\n\n") === "Rivi 3: 1. loppui ensin, 2. on ";
   *     vertaaString("33 hiljaa 1 hiipii\r\nhyv 33 tulee\r\n","33 hiljaa 1 hiipii\nhyv 33 tulee\n") === null;
   *     vertaaString("a b","a c") === "Ero riveiss 1: a b ja a c"
   *     vertaaString(null,null) === null;
   *     vertaaString(null," ") === "1. on null";
   *     vertaaString(" ",null) === "2. on null";
   * </pre>
   */
  public static String vertaaString(String ss1, String ss2)  { // NOPMD
    if ( ss1 == null & ss2 == null ) return null;
    if ( ss1 == null ) return "1. on null";
    if ( ss2 == null ) return "2. on null";
    ss1 = ss1.replaceAll("\\r\\n", "\n");
    ss2 = ss2.replaceAll("\\r\\n", "\n");
    ss1 = ss1.replaceAll("\\r", "\n");
    ss2 = ss2.replaceAll("\\r", "\n");  // Toki nyt voitaisiin verrata pelkki jonoja sellaisenaan
    Erottelija f1 = new Erottelija(ss1,"\n");
    Erottelija f2 = new Erottelija(ss2,"\n");
    int n = 1;
    while (true) {
      String s1 = null;
      String s2 = null;
      boolean b1 = f1.hasMoreTokens(); 
      boolean b2 = f2.hasMoreTokens(); 
      if ( b1 ) s1 = f1.nextToken();
      if ( b2 ) s2 = f2.nextToken();
      if ( !b1 && !b2)             return null;
      if (  b1 && !b2 )            return "Rivi "+ n + ": " + "2. loppui ensin, 1. on "+ s1;
      if ( !b1 &&  b2)             return "Rivi "+ n + ": " + "1. loppui ensin, 2. on "+ s2;
      if ( s1.compareTo(s2) != 0 ) return "Ero riveiss " + n + ": " + s1 + " ja " + s2;
      n++;
    }
  }

  /**
   * Kirjoitetaan tiedostoon sislt
   * @param nimi tiedoston nimi johon kirjoitetaan
   * @param sisalto merkkijono joka kirjoitetaan tiedostoon
   * @throws IOException jos tiedosto ei aukea.
   * @example
   * <pre name="test">
   * #THROWS IOException
   *     kirjoitaTiedosto("hiljaa1.txt", "33 hiljaa 1 hiipii\nhyv 33 tulee\n");
   *     tuhoaTiedosto("hiljaa1.txt");
   *     kirjoitaTiedosto(":\\.", "33 hiljaa 1 hiipii\nhyv 33 tulee\n"); #THROWS IOException
   * </pre>
   */
  public static void kirjoitaTiedosto(String nimi, String sisalto) throws IOException {
    PrintWriter out = null;
    try {
      out = new PrintWriter(new FileWriter(nimi));
      out.write(sisalto);
      out.close();
    } finally {
      if (out != null)
        out.close();
    }
  }

  /**
   * Tuhotaan tiedosto levylt
   * @param nimi tuhottavan tiedoston nimi
   */
  public static void tuhoaTiedosto(String nimi) {
    File f = new File(nimi);
    f.delete();
  }

  /**
   * Testataan tiedostojen vertaamista
   * @param args ei kytss
   * @throws IOException jos tulee virhe
   */
/*  
  public static void main(String[] args) throws IOException  {
    String koe = vertaaFileFile("fi/jyu/mit/ohj2/IO.java","fi/jyu/mit/ohj2/Syotto.java");
    System.out.println(koe);
  }
*/
}
